import '../ParserType.dart';
import '../ParserExceptions.dart';
import '../lexer/Lexer.dart';
import '../sub_types/UnconditionalAssignableSelector.dart';
import '../sub_types/Expressions.dart';
import '../main/Expression.dart';
import 'Identifier.dart';
import 'Literal.dart';

/// <primary> ::= <thisExpression>
/// | super <unconditionalAssignableSelector>
/// | <functionExpression>
/// | <literal>
/// | <identifier>
/// | <newExpression>
/// | <constObjectExpression>
/// | '(' <expression> ')'
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 74
class Primary extends ParserType {
  int caseNum;
  ThisExpression thisExpression;
  UnconditionalAssignableSelector unconditionalAssignableSelector;
  FunctionExpression functionExpression;
  Literal literal;
  Identifier identifier;
  NewExpression newExpression;
  ConstObjectExpression constObjectExpression;
  Expression expression;

  Primary(Lexer lexer) {
    int attempt, maxAttempts = 8;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
          case 2:
            _case2(lexer);
            break;
          case 3:
            _case3(lexer);
            break;
          case 4:
            _case4(lexer);
            break;
          case 5:
            _case5(lexer);
            break;
          case 6:
            _case6(lexer);
            break;
          case 7:
            _case7(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType('Tried all options in Primary: ALL FAILED');
    }
  }

  /// <thisExpression>
  void _case0(Lexer lexer) {
    thisExpression = ThisExpression(lexer);
    caseNum = 0;
  }

  /// super <unconditionalAssignableSelector>
  void _case1(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'super');
    lexer.advance();

    unconditionalAssignableSelector = UnconditionalAssignableSelector(lexer);
    caseNum = 1;
  }

  /// <functionExpression>
  void _case2(Lexer lexer) {
    functionExpression = FunctionExpression(lexer);
    caseNum = 2;
  }

  /// <literal>
  void _case3(Lexer lexer) {
    literal = Literal(lexer);
    caseNum = 3;
  }

  /// <identifier>
  void _case4(Lexer lexer) {
    identifier = Identifier(lexer);
    caseNum = 4;
  }

  /// <newExpression>
  void _case5(Lexer lexer) {
    newExpression = NewExpression(lexer);
    caseNum = 5;
  }

  /// <constObjectExpression>
  void _case6(Lexer lexer) {
    constObjectExpression = ConstObjectExpression(lexer);
    caseNum = 6;
  }

  /// '(' <expression> ')'
  void _case7(Lexer lexer) {
    checkType(lexer, TType.LEFT_PAREN);
    lexer.advance();
    expression = Expression(lexer);
    checkType(lexer, TType.RIGHT_PAREN);
    lexer.advance();
    caseNum = 7;
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return '$thisExpression';
        break;
      case 1:
        return 'super $unconditionalAssignableSelector';
        break;
      case 2:
        return '$functionExpression';
        break;
      case 3:
        return '$literal';
        break;
      case 4:
        return '$identifier';
        break;
      case 5:
        return '$newExpression';
        break;
      case 6:
        return '$constObjectExpression';
        break;
      case 7:
        return '($expression)';
        break;
    }
    return '';
  }
}
