import 'package:validators/validators.dart';
import 'CustomExceptions.dart';
import 'InputStream.dart';
import 'Token.dart';
import 'enumns.dart';

export 'enumns.dart';

class Lexer {
  List<Token> _tokenList;
  List<String> _keywords;
  InputStream _IS;
  State _state;
  String _soFar;
  int _startPos;
  int _lastRefPos;
  int _tokenIndex;
  bool _showWhitespace;

  Lexer() {
    _tokenList = [];
    _state = State.START;
    _soFar = "";
    _showWhitespace = false;
    _tokenIndex = 0;

    // Note that all keywords are assumed in all conditions.
    // 'https://www.dartlang.org/guides/language/language-tour' demonstrates different condtions, they are all assumed to always be true
    // EXCEPTIONS (have their own state): 'true', 'false'
    _keywords = [
      'else',
      'assert',
      'enum',
      'in',
      'super',
      'switch',
      'extends',
      'is',
      'break',
      'this',
      'case',
      'throw',
      'catch',
      'new',
      'class',
      'final',
      'null',
      'try',
      'const',
      'finally',
      'continue',
      'for',
      'var',
      'void',
      'default',
      'rethrow',
      'while',
      'return',
      'with',
      'do',
      'if', // These are keywords that are always identifiers
      'show', 'async', 'sync', 'on',
      'hide', // These are the keywords labeled with 3's on the dart language-tour
      'abstract',
      'dynamic',
      'implements',
      'as',
      'import',
      'static',
      'export',
      'interface',
      'external',
      'library',
      'factory',
      'mixin',
      'typedef',
      'operator',
      'covariant',
      'Function',
      'part',
      'get',
      'deferred',
      'set', // These are the keywords labeled with 2's on the language-tour
      'await', 'yield'
    ]; // These are the keywords labeled with 1's on the language-tour
  }

  Future<Lexer> parse(String fileName, [bool reset = true]) async {
    if (reset) {
      _tokenList = [];
      _state = State.START;
      _soFar = "";
      _tokenIndex = 0;
    }

    _IS = InputStream();
    await _IS.load(fileName);

    try {
      while (true) {
        _processChar(_IS.currentChar());
      }
    } on FormatException {
      // This means the lexer has finished parsing the file
      _add(TType.EOF);
    }

    return this;
  }

  Lexer parseSync(String fileName, [bool reset = true]) {
    if (reset) {
      _tokenList = [];
      _state = State.START;
      _soFar = "";
      _tokenIndex = 0;
    }

    _IS = InputStream();
    _IS.loadSync(fileName);

    try {
      while (true) {
        _processChar(_IS.currentChar());
      }
    } on FormatException {
      // This means the lexer has finished parsing the file
      _add(TType.EOF);
    }

    return this;
  }

  Lexer parseStr(String str, [bool reset = true]) {
    if (reset) {
      _tokenList = [];
      _state = State.START;
      _soFar = "";
      _tokenIndex = 0;
    }

    _IS = InputStream();
    _IS.loadStr(str);

    try {
      while (true) {
        _processChar(_IS.currentChar());
      }
    } on FormatException {
      // This means the lexer has finished parsing the file
      _add(TType.EOF);
    }

    return this;
  }

  void advance([int i = 1]) {
    _tokenIndex += i;
    if (_tokenIndex > numTokens) {
      throw EndOfFile('$currentIndex');
    }
  }

  void setPos(int i) => _tokenIndex = i;
  Token get currentToken => _tokenList[_tokenIndex];
  Token get previousToken => _tokenList[_tokenIndex - 1];
  Token get nextToken => _tokenList[_tokenIndex + 1];
  int get currentIndex => _tokenIndex;
  int get numTokens => _tokenList.length;
  operator [](int i) => _tokenList[i];

  String toString() {
    String rtn = '';
    if (_showWhitespace) {
      _tokenList.forEach((t) => rtn += t.toString() + '\n');
    } else {
      _tokenList.forEach((t) =>
          rtn += (t.type != TType.WHITESPACE ? t.toString() + '\n' : ''));
    }
    return rtn;
  }

  void _processChar(String c) {
    switch (_state) {
      case State.START:
        {
          _startState(c);
          break;
        }
      case State.SINGLE_LINE_COMMENT:
        {
          _singleLineCommentState(c);
          break;
        }
      case State.MULTI_LINE_COMMENT:
        {
          _multiLineCommentState(c);
          break;
        }
      case State.STRING:
        {
          _stringState(c);
          break;
        }
      case State.NUMBER:
        {
          _numberState(c);
          break;
        }
      case State.HEX_NUMBER:
        {
          _hexNumberState(c);
          break;
        }
      case State.UNDEFINED:
        {
          break;
        }
    }
    _IS.advance();
  }

  void _startState(String c) {
    _startPos = _IS.charLoc;
    if (isAlpha(c) || c == '_') {
      if (_soFar.isEmpty) {
        _lastRefPos = _startPos;
      }
      _soFar += c;
    } else if (isNumeric(c)) {
      if (_soFar.isEmpty) {
        _lastRefPos = _startPos;
        _state = State.NUMBER;
      }
      _soFar += c;
    } else if (c == '"' || c == "'") {
      // This is the string case.
      _state = State.STRING;
      _soFar += c;
    } else if (c == '/') {
      // This will start the comment case, or divide.
      if (_soFar.isNotEmpty) {
        _checkForWord();
      }
      _IS.advance();
      if (_IS.currentChar() == '/') {
        _state = State.SINGLE_LINE_COMMENT;
      } else if (_IS.currentChar() == '*') {
        _state = State.MULTI_LINE_COMMENT;
      } else {
        // is just a division
        _IS.stepBack();
        _soFar += c;
        _add(TType.DIVIDE);
      }
    } else {
      // Any thing inside here should not be allowed in an ID or KEYWORD.
      // In other words, it should be a single character, no state changes.
      _checkForWord();
      _soFar += c;

      if (c == '.') {
        _IS.advance();

        if (isNumeric(_IS.currentChar())) {
          _lastRefPos = _startPos;
          _state = State.NUMBER;
          _soFar += _IS.currentChar();
        } else {
          _IS.stepBack();
          _add(TType.PERIOD);
        }
      } else if (c.trim().isEmpty) {
        _pass(); // _add(TType.WHITESPACE); // No longer add whitespace tokens
      }
      else if (c == ',') {
        _add(TType.COMMA);
      }
      else if (c == ';') {
        _add(TType.SEMICOLON);
      }
      else if (c == ':') {
        _add(TType.COLON);
      }
      else if (c == '(') {
        _add(TType.LEFT_PAREN);
      }
      else if (c == ')') {
        _add(TType.RIGHT_PAREN);
      }
      else if (c == '[') {
        _add(TType.LEFT_BRACKET);
      }
      else if (c == ']') {
        _add(TType.RIGHT_BRACKET);
      }
      else if (c == '{') {
        _add(TType.LEFT_CURLY);
      }
      else if (c == '}') {
        _add(TType.RIGHT_CURLY);
      }
      else if (c == '<') {
        _add(TType.LEFT_ANGLE);
      }
      else if (c == '>') {
        _add(TType.RIGHT_ANGLE);
      }
      else if (c == '@') {
        _add(TType.KUKAC);
      }
      else if (c == '?') {
        _add(TType.Q_MARK);
      }
      else if (c == '+') {
        _add(TType.PLUS);
      }
      else if (c == '-') {
        _add(TType.MINUS);
      }
      else if (c == '*') {
        _add(TType.MULTIPLY);
      }
      else if (c == '=') {
        _add(TType.EQUALS);
      }
      else if (c == '|') {
        _add(TType.PIPE);
      }
      else if (c == '&') {
        _add(TType.AMPERSAND);
      }
      else if (c == '#') {
        _add(TType.HASH);
      }
      else if (c == '\$') {
        _add(TType.DOLLAR);
      }
      else if (c == '%') {
        _add(TType.PERCENT);
      }
      else if (c == '^') {
        _add(TType.CARAT);
      }
      else if (c == '!') {
        _add(TType.EXCLAMATION);
      }
      else if (c == '~') {
        _add(TType.TILDE);
      }
    }
  }

  void _checkForWord() {
    if (_soFar.isNotEmpty && (isAlpha(_soFar[0]) || _soFar[0] == '_')) {
      if (_keywords.contains(_soFar)) {
        _add(TType.KEYWORD);
      } else if (_soFar == 'true' || _soFar == 'false') {
        _add(TType.BOOLEAN);
      } else {
        _add(TType.IDENTIFIER);
      }
    }
  }

  void _singleLineCommentState(String c) {
    switch (c) {
      case '\n':
        // _add(TType.COMMENT);
        _pass();
        break;
      default:
        _soFar += c;
        break;
    }
  }

  void _multiLineCommentState(String c) {
    switch (c) {
      case '*':
        _IS.advance();
        if (_IS.currentChar() == '/') {
          _pass();
          // _add(TType.COMMENT);
        } else {
          _soFar += c;
          _soFar += _IS.currentChar();
        }
        break;
      default:
        _soFar += c;
        break;
    }
  }

  void _stringState(String c) {
    _soFar += c;
    if (c == '"' && _soFar[0] == '"') {
      _add(TType.STRING);
    } else if (c == "'" && _soFar[0] == "'") {
      _add(TType.STRING);
    } else if (c == '\\') {
      _IS.advance();
      // _soFar = _soFar.substring(0, _soFar.length - 1);
      _soFar += _IS.currentChar();
    }
  }

  void _numberState(String c) {
    if (_soFar == '0' && (c == 'x' || c == 'X')) {
      _soFar += c;
      _state = State.HEX_NUMBER;
    } else if (c == 'e' || c == 'E') {
      if (_soFar.contains('e') || _soFar.contains('E')) {
        throw LexerError('Found second instance of exponent in same number');
      }

      _soFar += c;
      _IS.advance();

      if (!(_IS.currentChar() == '+' || _IS.currentChar() == '-')) {
        throw LexerError('Expected "+" or "-", found "${_IS.currentChar()}"');
      }

      _soFar += _IS.currentChar();
    } else if (c == '.') {
      if (_soFar.contains('.')) {
        throw LexerError('Found second instance of period in same number');
      }

      _soFar += c;
    } else if (!isNumeric(_IS.currentChar())) {
      _IS.stepBack();
      _add(TType.NUMBER);
    } else {
      _soFar += c;
    }
  }

  void _hexNumberState(String c) {
    if (!(isNumeric(_IS.currentChar()) ||
        isAlpha(_IS.currentChar()) ||
        _IS.currentChar() == '_')) {
      _IS.stepBack();
      _add(TType.HEX_NUMBER);
    } else {
      _soFar += c;
    }
  }

  void _add(TType type) {
    int pos = (type == TType.IDENTIFIER ||
            type == TType.KEYWORD ||
            type == TType.NUMBER ||
            type == TType.HEX_NUMBER
        ? _lastRefPos
        : _startPos);
    _tokenList.add(Token(type, _soFar, pos));
    _state = State.START;
    _soFar = "";
  }

  void _pass() {
    _state = State.START;
    _soFar = "";
  }
}
