import '../ParserType.dart';
import '../ParserExceptions.dart';
import '../lexer/Lexer.dart';
import '../main/Expression.dart';
import '../sub_types/ForLoopParts.dart';

/// <elements> ::= <element> (',' <element>)* ','?
///
/// See https://github.com/dart-lang/language/blob/master/accepted/2.3/unified-collections/feature-specification.md
class Elements extends ParserType {
  List<Element> elements;
  bool hasComma;

  Elements(Lexer lexer) {
    elements = [];
    hasComma = false;

    elements.add(Element(lexer));

    int snapShot;
    try {
      while (isType(lexer, TType.COMMA)) {
        snapShot = lexer.currentIndex;
        lexer.advance();

        elements.add(Element(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    if (isType(lexer, TType.COMMA)) {
      lexer.advance();
      hasComma = true;
    }
  }

  String toString() {
    String rtn = '';
    elements.forEach((e) => rtn += ', $e');
    return rtn.substring(2) + (hasComma ? ',' : '');
  }

}

/// <element> ::= <expressionElement>
/// | <mapEntry>
/// | <spreadElement>
/// | <ifElement>
/// | <forElement>
///
/// See https://github.com/dart-lang/language/blob/master/accepted/2.3/unified-collections/feature-specification.md
class Element extends ParserType {
  int caseNum;
  ExpressionElement expressionElement;
  SpreadElement spreadElement;
  ForElement forElement;
  IfElement ifElement;
  MapEntry mapEntry;

  Element(Lexer lexer) {
    int attempt, maxAttempts = 5;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
          case 2:
            _case2(lexer);
            break;
          case 3:
            _case3(lexer);
            break;
          case 4:
            _case4(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType('Tried all options in Element: ALL FAILED');
    }
  }

  /// <mapEntry>
  void _case0(Lexer lexer) {
    mapEntry = MapEntry(lexer);
    caseNum = 0;
  }

  /// <expressionElement>
  void _case1(Lexer lexer) {
    expressionElement = ExpressionElement(lexer);
    caseNum = 1;
  }
  /// <spreadElement>
  void _case2(Lexer lexer) {
    spreadElement = SpreadElement(lexer);
    caseNum = 2;
  }

  /// <ifElement>
  void _case3(Lexer lexer) {
    ifElement = IfElement(lexer);
    caseNum = 3;
  }

  /// <forElement>
  void _case4(Lexer lexer) {
    forElement = ForElement(lexer);
    caseNum = 4;
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return '$mapEntry';
        break;
      case 1:
        return '$expressionElement';
        break;
      case 2:
        return '$spreadElement';
        break;
      case 3:
        return '$ifElement';
        break;
      case 4:
        return '$forElement';
        break;
    }
    return '';
  }
}

/// expressionElement ::= expression
/// 
/// See https://github.com/dart-lang/language/blob/master/accepted/2.3/unified-collections/feature-specification.md
class ExpressionElement extends ParserType {
  Expression expression;

  ExpressionElement(Lexer lexer) {
    expression = Expression(lexer);
  }

  String toString() {
    return '$expression';
  }

}


/// <mapEntry> ::= <expression> ':' <expression>
/// 
/// See https://github.com/dart-lang/language/blob/master/accepted/2.3/unified-collections/feature-specification.md
class MapEntry extends ParserType {
  Expression expression1;
  Expression expression2;

  MapEntry(Lexer lexer) {
    expression1 = Expression(lexer);

    checkType(lexer, TType.COLON);
    lexer.advance();

    expression2 = Expression(lexer);
  }

  String toString() {
    return '$expression1:$expression2';
  }

}

/// <spreadElement ::= ('...' | '...?') <expression>
/// 
/// See https://github.com/dart-lang/language/blob/master/accepted/2.3/unified-collections/feature-specification.md
class SpreadElement extends ParserType {
  Expression expression;
  bool isNullAware;

  SpreadElement(Lexer lexer) {
    isNullAware = false;

    checkType(lexer, TType.PERIOD);
    lexer.advance();
    checkType(lexer, TType.PERIOD);
    lexer.advance();
    checkType(lexer, TType.PERIOD);
    lexer.advance();

    if (isType(lexer, TType.Q_MARK)) {
      lexer.advance();
      isNullAware = true;
    }

    expression = Expression(lexer);
  }

  String toString() {
    return (isNullAware ? '...?' : '...') + '$expression';
  }

}

/// <ifElement> ::= 'if' '(' <expression> ')' <element> ( 'else' <element> )?
/// 
/// See https://github.com/dart-lang/language/blob/master/accepted/2.3/unified-collections/feature-specification.md
class IfElement extends ParserType {
  Expression expression;
  Element element1;
  Element element2;

  IfElement(Lexer lexer) {

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'if');
    lexer.advance();

    checkType(lexer, TType.LEFT_PAREN);
    lexer.advance();

    expression = Expression(lexer);

    checkType(lexer, TType.RIGHT_PAREN);
    lexer.advance();

    element1 = Element(lexer);

    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.KEYWORD);
      checkVal(lexer, 'else');
      lexer.advance();

      element2 = Element(lexer);
    }
    on WrongTokenType {
      lexer.setPos(snapShot);
    }
  }

  String toString() {
    return 'if ($expression) $element1' + (element2 != null ? ' else $element2' : '');
  }

}

/// <forElement> ::= 'await'? 'for' '(' <forLoopParts> ')' <element>
/// 
/// See https://github.com/dart-lang/language/blob/master/accepted/2.3/unified-collections/feature-specification.md
class ForElement extends ParserType {
  ForLoopParts forLoopParts;
  Element element;
  bool isAwait;

  ForElement(Lexer lexer) {

    isAwait = false;
    if (isType(lexer, TType.KEYWORD) && isVal(lexer, 'await')) {
      lexer.advance();
      isAwait = true;
    }

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'for');
    lexer.advance();

    checkType(lexer, TType.LEFT_PAREN);
    lexer.advance();

    forLoopParts = ForLoopParts(lexer);

    checkType(lexer, TType.RIGHT_PAREN);
    lexer.advance();

    element = Element(lexer);
  }

  /// <forElement> ::= 'await'? 'for' '(' <forLoopParts> ')' <element>
  String toString() {
    return (isAwait ? 'await' : '') + 'for ($forLoopParts) $element';
  }

}
