import '../ParserType.dart';
import '../ParserExceptions.dart';
import '../lexer/Lexer.dart';
import '../core/Identifier.dart';
import '../lists/TypeParameters.dart';
import '../main/MetaData.dart';
import '../main/SuperClass.dart';
import '../main/Interfaces.dart';
import '../main/ClassMemberDefinition.dart';
import '../main/MixinApplicationClass.dart';

/// <classDefinition> ::= <metadata> abstract? class <identifier> <typeParameters>?
/// <superclass>? <interfaces>?
/// '{' (<metadata> <classMemberDefinition>)* '}'
/// | <metadata> abstract? class <mixinApplicationClass>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 26

class ClassDefinition extends ParserType {
  int caseNum;
  List<MetaData> metaDatas;
  Identifier identifier;
  TypeParameters typeParameters;
  SuperClass superClass;
  Interfaces interfaces;
  List<ClassMemberDefinition> classMemberDefinitions;
  MixinApplicationClass mixinApplicationClass;
  bool isAbstract;

  ClassDefinition(Lexer lexer) {
    isAbstract = false;
    metaDatas = [];
    classMemberDefinitions = [];

    int attempt, maxAttempts = 2;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType('Tried all options in ClassDefinition: ALL FAILED');
    }
  }

  /// <metadata> abstract? class <identifier> <typeParameters>?
  /// <superclass>? <interfaces>?
  /// '{' (<metadata> <classMemberDefinition>)* '}'
  void _case0(Lexer lexer) {
    metaDatas.add(MetaData(lexer));

    if (isType(lexer, TType.KEYWORD) && isVal(lexer, 'abstract')) {
      isAbstract = true;
      lexer.advance();
    }

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'class');
    lexer.advance();

    identifier = Identifier(lexer);

    int snapShot = lexer.currentIndex;
    try {
      typeParameters = TypeParameters(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    snapShot = lexer.currentIndex;
    try {
      superClass = SuperClass(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    snapShot = lexer.currentIndex;
    try {
      interfaces = Interfaces(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    checkType(lexer, TType.LEFT_CURLY);
    lexer.advance();

    try {
      while (true) {
        snapShot = lexer.currentIndex;
        metaDatas.add(MetaData(lexer));
        classMemberDefinitions.add(ClassMemberDefinition(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    checkType(lexer, TType.RIGHT_CURLY);
    lexer.advance();

    caseNum = 0;
  }

  /// <metadata> abstract? class <mixinApplicationClass>
  void _case1(Lexer lexer) {
    metaDatas.add(MetaData(lexer));

    if (isType(lexer, TType.KEYWORD) && isVal(lexer, 'abstract')) {
      isAbstract = true;
      lexer.advance();
    }

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'class');
    lexer.advance();

    mixinApplicationClass = MixinApplicationClass(lexer);

    caseNum = 1;
  }

  String toString() {
    switch (caseNum) {
      case 0:
        String rtn = '${metaDatas[0]}' +
            (isAbstract ? 'abstract ' : '') +
            'class $identifier' +
            (typeParameters != null ? ' $typeParameters' : '') +
            (superClass != null ? ' $superClass' : '') +
            (interfaces != null ? ' $interfaces' : '') +
            ' { ';
        for (int i = 0; i < classMemberDefinitions.length; ++i) {
          rtn += '${metaDatas[i + 1]}${classMemberDefinitions[i]} ';
        }
        return rtn + '}';
        break;
      case 1:
        return '${metaDatas[0]}' +
            (isAbstract ? 'abstract ' : '') +
            'class $mixinApplicationClass';
        break;
    }
    return '';
  }
}
