import '../ParserType.dart';
import '../ParserExceptions.dart';
import '../lexer/Lexer.dart';
import '../sub_types/Expressions.dart';
import '../sub_types/Operators.dart';
import '../sub_types/CascadeSection.dart';

/// <expression> ::= <assignableExpression> <assignmentOperator> <expression>
/// | <conditionalExpression> <cascadeSection>*
/// | <throwExpression>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 74
class Expression extends ParserType {
  int caseNum;
  AssignableExpression assignableExpression;
  AssignmentOperator assignmentOperator;
  Expression expression;

  ConditionalExpression conditionalExpression;
  List<CascadeSection> cascadeSections;

  ThrowExpression throwExpression;

  Expression(Lexer lexer) {
    cascadeSections = [];
    // print ('---------------------------------------');
    // print ('Started Expression');
    // print (lexer.currentToken);
    // print (lexer.nextToken);

    int attempt, maxAttempts = 3;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 1:
            _case0(lexer);
            break;
          case 0:
            _case1(lexer);
            break;
          case 2:
            _case2(lexer);
            break;
        }

        // If it reaches here, a case succeded
        // print ('Expression succeeded $conditionalExpression $cascadeSections');
        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType('Tried all options in Expression: ALL FAILED');
    }

    // print ('Finished Expression');
  }

  /// <assignableExpression> <assignmentOperator> <expression>
  void _case0(Lexer lexer) {
    // print ('Starting Expression case0');
    assignableExpression = AssignableExpression(lexer);
    assignmentOperator = AssignmentOperator(lexer);
    expression = Expression(lexer);
    caseNum = 0;
    // print ('Completed Expression case0');
  }

  /// <conditionalExpression> <cascadeSection>*
  void _case1(Lexer lexer) {
    // print ('Starting Expression case1');
    conditionalExpression = ConditionalExpression(lexer);

    int snapShot;
    try {
      while (true) {
        snapShot = lexer.currentIndex;
        cascadeSections.add(CascadeSection(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    snapShot = lexer.currentIndex;
    AssignmentOperator tmp;
    try {
      tmp = AssignmentOperator(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    } finally {
      lexer.setPos(snapShot);
      if (tmp != null) {
        throw WrongTokenType('');
      }
    }

    caseNum = 1;

    // print ('Completed Expression case1');
  }

  /// <throwExpression>
  void _case2(Lexer lexer) {
    // print ('Starting Expression case2');
    throwExpression = ThrowExpression(lexer);
    caseNum = 2;
    // print ('Completed Expression case2');
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return '$assignableExpression $assignmentOperator $expression';
        break;
      case 1:
        String rtn = '$conditionalExpression';
        cascadeSections.forEach((s) => rtn += ' $s');
        return rtn;
        break;
      case 2:
        return '$throwExpression';
        break;
    }
    return '';
  }
}
