import '../ParserType.dart';
import '../ParserExceptions.dart';
import '../lexer/Lexer.dart';
import '../sub_types/TryParts.dart';
import '../main/Block.dart';

/// <tryStatement> ::= try <block> (<onPart>+ <finallyPart>? | <finallyPart>)
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 158
class TryStatement extends ParserType {
  Block block;
  List<OnPart> onParts;
  FinallyPart finallyPart;

  TryStatement(Lexer lexer) {
    onParts = [];

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'try');
    lexer.advance();

    block = Block(lexer);

    int attempt, maxAttempts = 2;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType('Tried all options in TryStatement: ALL FAILED');
    }
  }

  /// <onPart>+ <finallyPart>?
  void _case0(Lexer lexer) {
    onParts.add(OnPart(lexer));
    int snapShot;
    try {
      while (true) {
        snapShot = lexer.currentIndex;
        onParts.add(OnPart(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    snapShot = lexer.currentIndex;
    try {
      finallyPart = FinallyPart(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }
  }

  /// <finallyPart>
  void _case1(Lexer lexer) {
    finallyPart = FinallyPart(lexer);
  }

  String toString() {
    String rtn = 'try $block';
    if (onParts.isNotEmpty) {
      onParts.forEach((o) => rtn += ' $o');
      if (finallyPart != null) {
        rtn += ' $finallyPart';
      }
    } else if (finallyPart != null) {
      rtn += ' $finallyPart';
    }
    return rtn;
  }
}
