import '../ParserType.dart';
import '../ParserExceptions.dart';
import '../lexer/Lexer.dart';

import '../sub_types/Expressions.dart';
import '../sub_types/Operators.dart';
import 'AssignableSelector.dart';
import 'CascadeSelector.dart';
import 'ArgumentPart.dart';

/// <cascadeSection> ::= '..' (<cascadeSelector> <argumentPart>*)
/// (<assignableSelector> <argumentPart>*)*
/// (<assignmentOperator> <expressionWithoutCascade>)?
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 116
class CascadeSection extends ParserType {
  CascadeSelector cascadeSelector;
  List<ArgumentPart> argumentParts1;
  List<AssignableSelector> assignableSelectors;
  Map<int, List<ArgumentPart>> argumentParts2;
  AssignmentOperator assignmentOperator;
  ExpressionWithoutCascade expressionWithoutCascade;

  CascadeSection(Lexer lexer) {
    assignableSelectors = [];
    argumentParts1 = [];
    argumentParts2 = {};

    checkType(lexer, TType.PERIOD);
    lexer.advance();
    checkType(lexer, TType.PERIOD);
    lexer.advance();

    cascadeSelector = CascadeSelector(lexer);

    int snapShot;
    try {
      while (true) {
        snapShot = lexer.currentIndex;
        argumentParts1.add(ArgumentPart(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    try {
      int i = 0;
      while (true) {
        snapShot = lexer.currentIndex;
        assignableSelectors.add(AssignableSelector(lexer));

        argumentParts2[i] = [];
        int innerSnapShot;
        try {
          while (true) {
            innerSnapShot = lexer.currentIndex;
            argumentParts2[i].add(ArgumentPart(lexer));
          }
        } on WrongTokenType {
          lexer.setPos(innerSnapShot);
        }

        ++i;
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    snapShot = lexer.currentIndex;
    try {
      assignmentOperator = AssignmentOperator(lexer);
      expressionWithoutCascade = ExpressionWithoutCascade(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }
  }

  /// '..' (<cascadeSelector> <argumentPart>*)
  /// (<assignableSelector> <argumentPart>*)*
  /// (<assignmentOperator> <expressionWithoutCascade>)?
  String toString() {
    String rtn = '..$cascadeSelector';
    argumentParts1.forEach((p) => rtn += '$p');
    for (int i = 0; i < assignableSelectors.length; ++i) {
      rtn += '${assignableSelectors[i]}';
      argumentParts2[i].forEach((p) => rtn += '$p');
    }
    if (assignmentOperator != null) {
      rtn += ' $assignmentOperator $expressionWithoutCascade';
    }
    return rtn;
  }
}
