import '../ParserType.dart';
import '../ParserExceptions.dart';
import '../lexer/Lexer.dart';
import '../lists/StaticFinalDeclarationList.dart';
import '../lists/InitializedIdentifierList.dart';
import '../lists/Initializers.dart';
import '../sub_types/Signatures.dart';
import '../core/Type.dart';

/// <declaration> ::= <constantConstructorSignature> (<redirection> | <initializers>)?
/// | <constructorSignature> (<redirection> | <initializers>)?
/// | external <constantConstructorSignature>
/// | external <constructorSignature>
/// | (external static?)? <getterSignature>
/// | (external static?)? <setterSignature>
/// | external? <operatorSignature>
/// | (external static?)? <functionSignature>
/// | static (final | const) <type>? <staticFinalDeclarationList>
/// | final <type>? <initializedIdentifierList>
/// | (static | covariant)? (var | <type>) <initializedIdentifierList>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 26

class Declaration extends ParserType {
  int caseNum;

  ConstantConstructorSignature constantConstructorSignature;
  StaticFinalDeclarationList staticFinalDeclarationList;
  InitializedIdentifierList initializedIdentifierList;
  ConstructorSignature constructorSignature;
  OperatorSignature operatorSignature;
  FunctionSignature functionSignature;
  GetterSignature getterSignature;
  SetterSignature setterSignature;
  Initializers initializers;
  Redirection redirection;
  Type type;

  bool isCovariant;
  bool isExternal;
  bool isStatic;
  bool isConst;

  Declaration(Lexer lexer) {
    resetBools();

    int attempt, maxAttempts = 11;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
          case 2:
            _case2(lexer);
            break;
          case 3:
            _case3(lexer);
            break;
          case 4:
            _case4(lexer);
            break;
          case 5:
            _case5(lexer);
            break;
          case 6:
            _case6(lexer);
            break;
          case 7:
            _case7(lexer);
            break;
          case 8:
            _case8(lexer);
            break;
          case 9:
            _case9(lexer);
            break;
          case 10:
            _case10(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType('Tried all options in Declaration: ALL FAILED');
    }
  }

  void resetBools() {
    isCovariant = false;
    isExternal = false;
    isStatic = false;
    isConst = false;
  }

  /// <constantConstructorSignature> (<redirection> | <initializers>)?
  void _case0(Lexer lexer) {
    resetBools();

    constantConstructorSignature = ConstantConstructorSignature(lexer);

    int snapShot = lexer.currentIndex;
    try {
      int insideSnapShot = lexer.currentIndex;
      try {
        redirection = Redirection(lexer);
      } on WrongTokenType {
        lexer.setPos(insideSnapShot);

        initializers = Initializers(lexer);
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    caseNum = 0;
  }

  /// <constructorSignature> (<redirection> | <initializers>)?
  void _case1(Lexer lexer) {
    resetBools();

    constructorSignature = ConstructorSignature(lexer);

    int snapShot = lexer.currentIndex;
    try {
      int insideSnapShot = lexer.currentIndex;
      try {
        redirection = Redirection(lexer);
      } on WrongTokenType {
        lexer.setPos(insideSnapShot);

        initializers = Initializers(lexer);
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    caseNum = 1;
  }

  /// external <constantConstructorSignature>
  void _case2(Lexer lexer) {
    resetBools();

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'external');
    lexer.advance();

    constantConstructorSignature = ConstantConstructorSignature(lexer);

    caseNum = 2;
  }

  /// external <constructorSignature>
  void _case3(Lexer lexer) {
    resetBools();

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'external');
    lexer.advance();

    constructorSignature = ConstructorSignature(lexer);

    caseNum = 3;
  }

  /// (external static?)? <getterSignature>
  void _case4(Lexer lexer) {
    resetBools();

    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.KEYWORD);
      checkVal(lexer, 'external');
      lexer.advance();

      isExternal = true;
      snapShot = lexer.currentIndex;

      checkType(lexer, TType.KEYWORD);
      checkVal(lexer, 'static');
      lexer.advance();

      isStatic = true;
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    getterSignature = GetterSignature(lexer);

    caseNum = 4;
  }

  /// (external static?)? <setterSignature>
  void _case5(Lexer lexer) {
    resetBools();

    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.KEYWORD);
      checkVal(lexer, 'external');
      lexer.advance();

      isExternal = true;
      snapShot = lexer.currentIndex;

      checkType(lexer, TType.KEYWORD);
      checkVal(lexer, 'static');
      lexer.advance();

      isStatic = true;
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    setterSignature = SetterSignature(lexer);

    caseNum = 5;
  }

  /// external? <operatorSignature>
  void _case6(Lexer lexer) {
    resetBools();

    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.KEYWORD);
      checkVal(lexer, 'external');
      lexer.advance();

      isExternal = true;
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    operatorSignature = OperatorSignature(lexer);

    caseNum = 6;
  }

  /// (external static?)? <functionSignature>
  void _case7(Lexer lexer) {
    resetBools();

    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.KEYWORD);
      checkVal(lexer, 'external');
      lexer.advance();

      isExternal = true;
      snapShot = lexer.currentIndex;

      checkType(lexer, TType.KEYWORD);
      checkVal(lexer, 'static');
      lexer.advance();

      isStatic = true;
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    functionSignature = FunctionSignature(lexer);

    caseNum = 7;
  }

  /// static (final | const) <type>? <staticFinalDeclarationList>
  void _case8(Lexer lexer) {
    resetBools();

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'static');
    lexer.advance();

    if (isType(lexer, TType.KEYWORD) && isVal(lexer, 'final')) {
      isConst = false;
    } else if (isType(lexer, TType.KEYWORD) && isVal(lexer, 'const')) {
      isConst = true;
    } else {
      throw WrongTokenType('Expected "final" or "const"');
    }
    lexer.advance();

    int snapShot = lexer.currentIndex;
    try {
      type = Type(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    staticFinalDeclarationList = StaticFinalDeclarationList(lexer);

    caseNum = 8;
  }

  /// final <type>? <initializedIdentifierList>
  void _case9(Lexer lexer) {
    resetBools();

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'final');
    lexer.advance();

    int snapShot = lexer.currentIndex;
    try {
      type = Type(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    initializedIdentifierList = InitializedIdentifierList(lexer);

    caseNum = 9;
  }

  /// (static | covariant)? (var | <type>) <initializedIdentifierList>
  void _case10(Lexer lexer) {
    resetBools();

    if (isType(lexer, TType.KEYWORD) && isVal(lexer, 'static')) {
      isStatic = true;
      lexer.advance();
    } else if (isType(lexer, TType.KEYWORD) && isVal(lexer, 'covariant')) {
      isCovariant = true;
      lexer.advance();
    }

    if (isType(lexer, TType.KEYWORD) && isVal(lexer, 'var')) {
      lexer.advance();
    } else {
      type = Type(lexer);
    }

    initializedIdentifierList = InitializedIdentifierList(lexer);

    caseNum = 10;
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return '$constantConstructorSignature' +
            (redirection != null ? ' $redirection' : ' $initializers');
        break;
      case 1:
        return '$constructorSignature' +
            (redirection != null ? ' $redirection' : '') +
            (initializers != null ? ' $initializers' : '');
        break;
      case 2:
        return 'external $constantConstructorSignature';
        break;
      case 3:
        return 'external $constructorSignature';
        break;
      case 4:
        return (isExternal ? 'external ' : '') +
            (isStatic ? 'static ' : '') +
            '$getterSignature';
        break;
      case 5:
        return (isExternal ? 'external ' : '') +
            (isStatic ? 'static ' : '') +
            '$setterSignature';
        break;
      case 6:
        return (isExternal ? 'external ' : '') + '$operatorSignature';
        break;
      case 7:
        return (isExternal ? 'external ' : '') +
            (isStatic ? 'static ' : '') +
            '$functionSignature';
        break;
      case 8:
        return 'static ' +
            (isConst ? 'const ' : 'final ') +
            (type != null ? '$type ' : '') +
            '$staticFinalDeclarationList';
        break;
      case 9:
        return 'final ' +
            (type != null ? '$type ' : '') +
            '$initializedIdentifierList';
        break;
      case 10:
        return (isCovariant ? 'covariant ' : '') +
            (isStatic ? 'static ' : '') +
            (type != null ? '$type ' : 'var ') +
            '$initializedIdentifierList';
        break;
    }
    return '';
  }
}
