import '../ParserExceptions.dart';
import '../ParserType.dart';
import '../lexer/Lexer.dart';
import '../lists/Arguments.dart';
import '../core/Primary.dart';
import '../core/Identifier.dart';
import '../main/Selector.dart';
import '../main/Expression.dart';
import '../sub_types/Types.dart';
import '../sub_types/FunctionBody.dart';
import '../sub_types/FormalParameterPart.dart';
import 'UnconditionalAssignableSelector.dart';
import 'AssignableSelectorPart.dart';
import 'ConstructorInvocation.dart';
import 'Operators.dart';

/// <expressionWithoutCascade> ::=
/// <assignableExpression> <assignmentOperator> <expressionWithoutCascade>
/// | <conditionalExpression>
/// | <throwExpressionWithoutCascade>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 74
class ExpressionWithoutCascade extends ParserType {
  int caseNum;
  AssignableExpression assignableExpression;
  AssignmentOperator assignmentOperator;
  ExpressionWithoutCascade expressionWithoutCascade;

  ConditionalExpression conditionalExpression;

  ThrowExpressionWithoutCascade throwExpressionWithoutCascade;

  ExpressionWithoutCascade(Lexer lexer) {
    int attempt, maxAttempts = 3;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
          case 2:
            _case2(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType(
          'Tried all options in ExpressionWithoutCascade: ALL FAILED');
    }
  }

  /// <assignableExpression> <assignmentOperator> <expressionWithoutCascade>
  void _case0(Lexer lexer) {
    assignableExpression = AssignableExpression(lexer);
    assignmentOperator = AssignmentOperator(lexer);
    expressionWithoutCascade = ExpressionWithoutCascade(lexer);
    caseNum = 0;
  }

  /// <conditionalExpression>
  void _case1(Lexer lexer) {
    conditionalExpression = ConditionalExpression(lexer);
    caseNum = 1;
  }

  /// <throwExpressionWithoutCascade>
  void _case2(Lexer lexer) {
    throwExpressionWithoutCascade = ThrowExpressionWithoutCascade(lexer);
    caseNum = 2;
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return '$assignableExpression $assignmentOperator $expressionWithoutCascade';
        break;
      case 1:
        return '$conditionalExpression';
        break;
      case 2:
        return '$throwExpressionWithoutCascade';
        break;
    }
    return '';
  }
}

/// <throwExpression> ::= throw <expression>
///
/// Page 93
class ThrowExpression extends ParserType {
  Expression expression;

  ThrowExpression(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'throw');
    lexer.advance();

    expression = Expression(lexer);
  }

  String toString() {
    return 'throw $expression';
  }
}

/// <functionExpression> ::= <formalParameterPart> <functionBody>
/// MODIFIED: SPECIAL CASE ALLOWS FUNCTIONBODY WITH NO ';' with =>
class FunctionExpression extends ParserType {
  FormalParameterPart formalParameterPart;
  FunctionBody functionBody;

  FunctionExpression(Lexer lexer) {
    formalParameterPart = FormalParameterPart(lexer);
    functionBody = FunctionBody(lexer, false);
  }

  String toString() {
    return '$formalParameterPart $functionBody';
  }
}

/// <throwExpressionWithoutCascade> ::= throw <expressionWithoutCascade>
class ThrowExpressionWithoutCascade extends ParserType {
  ExpressionWithoutCascade expressionWithoutCascade;

  ThrowExpressionWithoutCascade(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'throw');
    lexer.advance();

    expressionWithoutCascade = ExpressionWithoutCascade(lexer);
  }
}

/// <conditionalExpression> ::= <ifNullExpression>
/// ('?' <expressionWithoutCascade> ':' <expressionWithoutCascade>)?
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 130-140
class ConditionalExpression extends ParserType {
  IfNullExpression ifNullExpression;
  ExpressionWithoutCascade expressionWithoutCascade1;
  ExpressionWithoutCascade expressionWithoutCascade2;

  ConditionalExpression(Lexer lexer) {
    ifNullExpression = IfNullExpression(lexer);

    int snapShot = lexer.currentIndex;
    try {
      if (isType(lexer, TType.Q_MARK)) {
        lexer.advance();
        expressionWithoutCascade1 = ExpressionWithoutCascade(lexer);
        checkType(lexer, TType.COLON);
        lexer.advance();
        expressionWithoutCascade2 = ExpressionWithoutCascade(lexer);
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }
  }

  String toString() {
    return '$ifNullExpression' +
        (expressionWithoutCascade1 != null
            ? ' ? $expressionWithoutCascade1 : $expressionWithoutCascade2'
            : '');
  }
}

/// <ifNullExpression> ::= <logicalOrExpression> ('??' <logicalOrExpression>)*
class IfNullExpression extends ParserType {
  List<LogicalOrExpression> logicalOrExpressions;

  IfNullExpression(Lexer lexer) {
    logicalOrExpressions = [];
    logicalOrExpressions.add(LogicalOrExpression(lexer));

    int snapShot;
    try {
      while (true) {
        snapShot = lexer.currentIndex;
        checkType(lexer, TType.Q_MARK);
        lexer.advance();
        checkType(lexer, TType.Q_MARK);
        lexer.advance();

        logicalOrExpressions.add(LogicalOrExpression(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }
  }

  String toString() {
    String rtn = '';
    logicalOrExpressions.forEach((e) => rtn += '?? $e ');
    return rtn.substring(3).trimRight();
  }
}

/// <logicalOrExpression> ::= <logicalAndExpression> ('||' <logicalAndExpression>)*
class LogicalOrExpression extends ParserType {
  List<LogicalAndExpression> logicalAndExpressions;

  LogicalOrExpression(Lexer lexer) {
    logicalAndExpressions = [];
    logicalAndExpressions.add(LogicalAndExpression(lexer));

    int snapShot;
    try {
      while (true) {
        snapShot = lexer.currentIndex;
        checkType(lexer, TType.PIPE);
        lexer.advance();
        checkType(lexer, TType.PIPE);
        lexer.advance();

        logicalAndExpressions.add(LogicalAndExpression(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }
  }

  String toString() {
    String rtn = '';
    logicalAndExpressions.forEach((e) => rtn += '|| $e ');
    return rtn.substring(3).trimRight();
  }
}

/// <logicalAndExpression> ::= <equalityExpression> ('&&' <equalityExpression>)*
class LogicalAndExpression extends ParserType {
  List<EqualityExpression> equalityExpressions;

  LogicalAndExpression(Lexer lexer) {
    equalityExpressions = [];
    equalityExpressions.add(EqualityExpression(lexer));

    int snapShot;
    try {
      while (true) {
        snapShot = lexer.currentIndex;
        checkType(lexer, TType.AMPERSAND);
        lexer.advance();
        checkType(lexer, TType.AMPERSAND);
        lexer.advance();

        equalityExpressions.add(EqualityExpression(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }
  }

  String toString() {
    String rtn = '';
    equalityExpressions.forEach((e) => rtn += '&& $e ');
    return rtn.substring(3).trimRight();
  }
}

/// <equalityExpression> ::= <relationalExpression> (<equalityOperator> <relationalExpression>)?
/// | super <equalityOperator> <relationalExpression>
class EqualityExpression extends ParserType {
  RelationalExpression relationalExpression1;
  RelationalExpression relationalExpression2;
  EqualityOperator equalityOperator;
  bool isSuper;

  EqualityExpression(Lexer lexer) {
    int attempt, maxAttempts = 2;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType(
          'Tried all options in EqualityExpression: ALL FAILED');
    }
  }

  /// <relationalExpression> (<equalityOperator> <relationalExpression>)?
  void _case0(Lexer lexer) {
    relationalExpression1 = RelationalExpression(lexer);

    int snapShot = lexer.currentIndex;
    try {
      equalityOperator = EqualityOperator(lexer);
      relationalExpression2 = RelationalExpression(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    isSuper = false;
  }

  /// super <equalityOperator> <relationalExpression>
  void _case1(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'super');
    lexer.advance();

    equalityOperator = EqualityOperator(lexer);
    relationalExpression1 = RelationalExpression(lexer);

    isSuper = true;
  }

  String toString() {
    if (isSuper) {
      return 'super $equalityOperator $relationalExpression1';
    } else {
      return '$relationalExpression1'.trimRight() +
          (equalityOperator != null
                  ? ' $equalityOperator $relationalExpression2'
                  : '')
              .trimRight();
    }
  }
}

/// <relationalExpression> ::= <bitwiseOrExpression>
/// (<typeTest> | <typeCast> | <relationalOperator> <bitwiseOrExpression>)?
/// | super <relationalOperator> <bitwiseOrExpression>
class RelationalExpression extends ParserType {
  int insideCaseNum;
  BitwiseOrExpression bitwiseOrExpression1;
  BitwiseOrExpression bitwiseOrExpression2;
  RelationalOperator relationalOperator;
  TypeTest typeTest;
  TypeCast typeCast;
  bool isSuper;

  RelationalExpression(Lexer lexer) {
    int attempt, maxAttempts = 2;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType(
          'Tried all options in RelationalExpression: ALL FAILED');
    }
  }

  /// <bitwiseOrExpression>
  /// (<typeTest> | <typeCast> | <relationalOperator> <bitwiseOrExpression>)?
  void _case0(Lexer lexer) {
    bitwiseOrExpression1 = BitwiseOrExpression(lexer);

    int attempt, maxAttempts = 3;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            typeTest = TypeTest(lexer);
            insideCaseNum = 0;
            break;
          case 1:
            typeCast = TypeCast(lexer);
            insideCaseNum = 1;
            break;
          case 2:
            relationalOperator = RelationalOperator(lexer);
            bitwiseOrExpression2 = BitwiseOrExpression(lexer);
            insideCaseNum = 2;
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      insideCaseNum = 3;
    }

    isSuper = false;
  }

  /// super <relationalOperator> <bitwiseOrExpression>
  void _case1(Lexer lexer) {
    relationalOperator = RelationalOperator(lexer);
    bitwiseOrExpression1 = BitwiseOrExpression(lexer);

    isSuper = true;
  }

  String toString() {
    if (isSuper) {
      return 'super $relationalOperator $bitwiseOrExpression1';
    } else {
      String rtn = '$bitwiseOrExpression1';
      switch (insideCaseNum) {
        case 0:
          rtn += ' $typeTest';
          break;
        case 1:
          rtn += ' $typeCast';
          break;
        case 2:
          rtn += ' $relationalOperator $bitwiseOrExpression2';
      }
      return rtn;
    }
  }
}

/// <bitwiseOrExpression> ::=
/// <bitwiseXorExpression> ('|' <bitwiseXorExpression>)*
/// | super ('|' <bitwiseXorExpression>)+
class BitwiseOrExpression extends ParserType {
  List<BitwiseXorExpression> bitwiseXorExpressions;
  bool isSuper;

  BitwiseOrExpression(Lexer lexer) {
    bitwiseXorExpressions = [];

    int attempt, maxAttempts = 2;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType(
          'Tried all options in BitwiseOrExpression: ALL FAILED');
    }
  }

  /// <bitwiseXorExpression> ('|' <bitwiseXorExpression>)*
  void _case0(Lexer lexer) {
    bitwiseXorExpressions.add(BitwiseXorExpression(lexer));

    int snapShot;
    try {
      while (isType(lexer, TType.PIPE)) {
        snapShot = lexer.currentIndex;
        lexer.advance();
        bitwiseXorExpressions.add(BitwiseXorExpression(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    isSuper = false;
  }

  /// super ('|' <bitwiseXorExpression>)+
  void _case1(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'super');
    lexer.advance();

    int snapShot;
    try {
      while (isType(lexer, TType.PIPE)) {
        snapShot = lexer.currentIndex;
        lexer.advance();
        bitwiseXorExpressions.add(BitwiseXorExpression(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    isSuper = true;
  }

  String toString() {
    String rtn = '';
    if (isSuper) {
      rtn += 'super';
      bitwiseXorExpressions.forEach((e) => rtn += ' | $e');
    } else {
      bitwiseXorExpressions.forEach((e) => rtn += '$e | ');
      rtn = rtn.substring(0, rtn.length - 3);
    }
    return rtn;
  }
}

/// <bitwiseXorExpression> ::=
/// <bitwiseAndExpression> ('^' <bitwiseAndExpression>)*
/// | super ('^' <bitwiseAndExpression>)+
class BitwiseXorExpression extends ParserType {
  List<BitwiseAndExpression> bitwiseAndExpressions;
  bool isSuper;

  BitwiseXorExpression(Lexer lexer) {
    bitwiseAndExpressions = [];

    int attempt, maxAttempts = 2;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType(
          'Tried all options in BitwiseXorExpression: ALL FAILED');
    }
  }

  /// <bitwiseAndExpression> ('^' <bitwiseAndExpression>)*
  void _case0(Lexer lexer) {
    bitwiseAndExpressions.add(BitwiseAndExpression(lexer));

    int snapShot;
    try {
      while (isType(lexer, TType.CARAT)) {
        snapShot = lexer.currentIndex;
        lexer.advance();
        bitwiseAndExpressions.add(BitwiseAndExpression(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    isSuper = false;
  }

  /// super ('^' <bitwiseAndExpression>)+
  void _case1(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'super');
    lexer.advance();

    int snapShot;
    try {
      while (isType(lexer, TType.CARAT)) {
        snapShot = lexer.currentIndex;
        lexer.advance();
        bitwiseAndExpressions.add(BitwiseAndExpression(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    isSuper = true;
  }

  String toString() {
    String rtn = '';
    if (isSuper) {
      rtn += 'super';
      bitwiseAndExpressions.forEach((e) => rtn += ' ^ $e');
    } else {
      bitwiseAndExpressions.forEach((e) => rtn += '$e ^ ');
      rtn = rtn.substring(0, rtn.length - 3);
    }
    return rtn;
  }
}

/// <bitwiseAndExpression> ::= <shiftExpression> ('&' <shiftExpression>)*
/// | super ('&' <shiftExpression>)+
class BitwiseAndExpression extends ParserType {
  List<ShiftExpression> shiftExpressions;
  bool isSuper;

  BitwiseAndExpression(Lexer lexer) {
    shiftExpressions = [];

    int attempt, maxAttempts = 2;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType(
          'Tried all options in BitwiseAndExpression: ALL FAILED');
    }
  }

  /// <shiftExpression> ('&' <shiftExpression>)*
  void _case0(Lexer lexer) {
    shiftExpressions.add(ShiftExpression(lexer));

    int snapShot;
    try {
      while (isType(lexer, TType.AMPERSAND)) {
        snapShot = lexer.currentIndex;
        lexer.advance();
        shiftExpressions.add(ShiftExpression(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    isSuper = false;
  }

  /// super ('&' <shiftExpression>)+
  void _case1(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'super');
    lexer.advance();

    int snapShot;
    try {
      while (isType(lexer, TType.AMPERSAND)) {
        snapShot = lexer.currentIndex;
        lexer.advance();
        shiftExpressions.add(ShiftExpression(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    isSuper = true;
  }

  String toString() {
    String rtn = '';
    if (isSuper) {
      rtn += 'super';
      shiftExpressions.forEach((e) => rtn += ' & $e');
    } else {
      shiftExpressions.forEach((e) => rtn += '$e & ');
      rtn = rtn.substring(0, rtn.length - 3);
    }
    return rtn;
  }
}

/// <shiftExpression> ::=
/// <additiveExpression> (<shiftOperator> <additiveExpression>)*
/// | super (<shiftOperator> <additiveExpression>)+
class ShiftExpression extends ParserType {
  List<ShiftOperator> shiftOperators;
  List<AdditiveExpression> additiveExpressions;
  bool isSuper;

  ShiftExpression(Lexer lexer) {
    shiftOperators = [];
    additiveExpressions = [];

    int attempt, maxAttempts = 2;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType('Tried all options in ShiftExpression: ALL FAILED');
    }
  }

  /// <additiveExpression> (<shiftOperator> <additiveExpression>)*
  void _case0(Lexer lexer) {
    additiveExpressions.add(AdditiveExpression(lexer));

    int snapShot;
    try {
      while (true) {
        snapShot = lexer.currentIndex;
        shiftOperators.add(ShiftOperator(lexer));
        additiveExpressions.add(AdditiveExpression(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    isSuper = false;
  }

  /// super (<shiftOperator> <additiveExpression>)+
  void _case1(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'super');
    lexer.advance();

    shiftOperators.add(ShiftOperator(lexer));
    additiveExpressions.add(AdditiveExpression(lexer));

    int snapShot;
    try {
      while (true) {
        snapShot = lexer.currentIndex;
        shiftOperators.add(ShiftOperator(lexer));
        additiveExpressions.add(AdditiveExpression(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    isSuper = true;
  }

  String toString() {
    String rtn = '';
    if (isSuper) {
      rtn += 'super';
      for (int i = 0; i < shiftOperators.length; ++i) {
        rtn += ' ${shiftOperators[i]} ${additiveExpressions[i]}';
      }
    } else {
      rtn += '${additiveExpressions[0]}';
      for (int i = 0; i < shiftOperators.length; ++i) {
        rtn += ' ${shiftOperators[i]} ${additiveExpressions[i + 1]}';
      }
    }
    return rtn;
  }
}

/// <additiveExpression> ::= <multiplicativeExpression>
/// (<additiveOperator> <multiplicativeExpression>)*
/// | super (<additiveOperator> <multiplicativeExpression>)+
class AdditiveExpression extends ParserType {
  List<AdditiveOperator> additiveOperators;
  List<MultiplicativeExpression> multiplicativeExpressions;
  bool isSuper;

  AdditiveExpression(Lexer lexer) {
    additiveOperators = [];
    multiplicativeExpressions = [];

    int attempt, maxAttempts = 2;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType(
          'Tried all options in AdditiveExpression: ALL FAILED');
    }
  }

  /// <multiplicativeExpression> (<additiveOperator> <multiplicativeExpression>)*
  void _case0(Lexer lexer) {
    multiplicativeExpressions.add(MultiplicativeExpression(lexer));

    int snapShot;
    try {
      while (true) {
        snapShot = lexer.currentIndex;
        additiveOperators.add(AdditiveOperator(lexer));
        multiplicativeExpressions.add(MultiplicativeExpression(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    isSuper = false;
  }

  /// super (<additiveOperator> <multiplicativeExpression>)+
  void _case1(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'super');
    lexer.advance();

    additiveOperators.add(AdditiveOperator(lexer));
    multiplicativeExpressions.add(MultiplicativeExpression(lexer));

    int snapShot;
    try {
      while (true) {
        snapShot = lexer.currentIndex;
        additiveOperators.add(AdditiveOperator(lexer));
        multiplicativeExpressions.add(MultiplicativeExpression(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    isSuper = true;
  }

  String toString() {
    String rtn = '';
    if (isSuper) {
      rtn += 'super';
      for (int i = 0; i < additiveOperators.length; ++i) {
        rtn += ' ${additiveOperators[i]} ${multiplicativeExpressions[i]}';
      }
    } else {
      rtn += '${multiplicativeExpressions[0]}';
      for (int i = 0; i < additiveOperators.length; ++i) {
        rtn += ' ${additiveOperators[i]} ${multiplicativeExpressions[i + 1]}';
      }
    }
    return rtn;
  }
}

/// <multiplicativeExpression> ::=
/// <unaryExpression> (<multiplicativeOperator> <unaryExpression>)*
/// | super (<multiplicativeOperator> <unaryExpression>)+
class MultiplicativeExpression extends ParserType {
  List<MultiplicativeOperator> multiplicativeOperators;
  List<UnaryExpression> unaryExpressions;
  bool isSuper;

  MultiplicativeExpression(Lexer lexer) {
    multiplicativeOperators = [];
    unaryExpressions = [];

    int attempt, maxAttempts = 2;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType(
          'Tried all options in MultiplicativeExpression: ALL FAILED');
    }
  }

  /// <unaryExpression> (<multiplicativeOperator> <unaryExpression>)*
  void _case0(Lexer lexer) {
    unaryExpressions.add(UnaryExpression(lexer));

    int snapShot;
    try {
      while (true) {
        snapShot = lexer.currentIndex;
        multiplicativeOperators.add(MultiplicativeOperator(lexer));
        unaryExpressions.add(UnaryExpression(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    isSuper = false;
  }

  /// super (<multiplicativeOperator> <unaryExpression>)+
  void _case1(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'super');
    lexer.advance();

    multiplicativeOperators.add(MultiplicativeOperator(lexer));
    unaryExpressions.add(UnaryExpression(lexer));

    int snapShot;
    try {
      while (true) {
        snapShot = lexer.currentIndex;
        multiplicativeOperators.add(MultiplicativeOperator(lexer));
        unaryExpressions.add(UnaryExpression(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    isSuper = true;
  }

  String toString() {
    String rtn = '';
    if (isSuper) {
      rtn += 'super';
      for (int i = 0; i < multiplicativeOperators.length; ++i) {
        rtn += ' ${multiplicativeOperators[i]} ${unaryExpressions[i]}';
      }
    } else {
      rtn += '${unaryExpressions[0]}';
      for (int i = 0; i < multiplicativeOperators.length; ++i) {
        rtn += ' ${multiplicativeOperators[i]} ${unaryExpressions[i + 1]}';
      }
    }
    return rtn;
  }
}

/// <unaryExpression> ::= <prefixOperator> <unaryExpression>
/// | <awaitExpression>
/// | <postfixExpression>
/// | (<minusOperator> | <tildeOperator>) super
/// | <incrementOperator> <assignableExpression>
class UnaryExpression extends ParserType {
  int caseNum;
  PrefixOperator prefixOperator;
  UnaryExpression unaryExpression;

  AwaitExpression awaitExpression;

  PostfixExpression postfixExpression;

  MinusOperator minusOperator;
  TildeOperator tildeOperator;

  IncrementOperator incrementOperator;
  AssignableExpression assignableExpression;

  UnaryExpression(Lexer lexer) {
    // print ('Starting UnaryExpression');
    // print (lexer.currentToken);
    // print (lexer.nextToken);
    int attempt, maxAttempts = 5;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
          case 2:
            _case2(lexer);
            break;
          case 3:
            _case3(lexer);
            break;
          case 4:
            _case4(lexer);
            break;
        }

        // print ('Completed UnaryExpression $caseNum');
        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      // print ('Failed UnaryExpression');
      throw WrongTokenType('Tried all options in UnaryExpression: ALL FAILED');
    }
  }

  /// <prefixOperator> <unaryExpression>
  void _case0(Lexer lexer) {
    // print ('Starting UnaryExpression case0');
    prefixOperator = PrefixOperator(lexer);
    unaryExpression = UnaryExpression(lexer);
    caseNum = 0;
    // print ('Completed UnaryExpression case0');
  }

  /// <awaitExpression>
  void _case1(Lexer lexer) {
    // print ('Starting UnaryExpression case1');
    awaitExpression = AwaitExpression(lexer);
    caseNum = 1;
    // print ('Completed UnaryExpression case1');
  }

  /// <postfixExpression>
  void _case2(Lexer lexer) {
    // print ('Starting UnaryExpression case2');
    postfixExpression = PostfixExpression(lexer);
    caseNum = 2;
    // print ('Completed UnaryExpression case2');
  }

  /// (<minusOperator> | <tildeOperator>) super
  void _case3(Lexer lexer) {
    // print ('Starting UnaryExpression case3');
    int snapShot = lexer.currentIndex;
    try {
      minusOperator = MinusOperator(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);

      tildeOperator = TildeOperator(lexer);
    }

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'super');
    lexer.advance();
    caseNum = 3;
    // print ('Completed UnaryExpression case3');
  }

  /// <incrementOperator> <assignableExpression>
  void _case4(Lexer lexer) {
    // print ('Starting UnaryExpression case4');
    incrementOperator = IncrementOperator(lexer);
    assignableExpression = AssignableExpression(lexer);
    caseNum = 4;
    // print ('Completed UnaryExpression case4');
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return '$prefixOperator$unaryExpression';
        break;
      case 1:
        return '$awaitExpression';
        break;
      case 2:
        ;
        return '$postfixExpression';
        break;
      case 3:
        if (minusOperator != null) {
          return '$minusOperator' + 'super';
        } else {
          return '$tildeOperator' + 'super';
        }
        break;
      case 4:
        return '$incrementOperator$assignableExpression';
        break;
    }
    return '';
  }
}

/// <awaitExpression> ::= await <unaryExpression>
class AwaitExpression extends ParserType {
  UnaryExpression unaryExpression;

  AwaitExpression(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'await');
    lexer.advance();

    unaryExpression = UnaryExpression(lexer);
  }

  String toString() {
    return 'await $unaryExpression';
  }
}

/// <postfixExpression> ::= <assignableExpression> <postfixOperator>
/// | <constructorInvocation> <selector>*
/// | <primary> <selector>*
class PostfixExpression extends ParserType {
  int caseNum;
  AssignableExpression assignableExpression;
  PostfixOperator posfixOperator;
  ConstructorInvocation constructorInvocation;
  Primary primary;
  List<Selector> selectors;

  PostfixExpression(Lexer lexer) {
    // print ('Starting PostfixExpression');
    // print (lexer.currentToken);
    // print (lexer.nextToken);
    int attempt, maxAttempts = 3;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 2:
            _case0(lexer);
            break;
          case 0:
            _case1(lexer);
            break;
          case 1:
            _case2(lexer);
            break;
        }

        // print ('Completed PostfixExpression $caseNum');
        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType('Tried all options in PosfixExpression: ALL FAILED');
    }
  }

  /// <assignableExpression> <postfixOperator>
  void _case0(Lexer lexer) {
    // print ('Starting PostfixExpression case0');
    // print (lexer.currentToken);
    assignableExpression = AssignableExpression(lexer);
    posfixOperator = PostfixOperator(lexer);
    caseNum = 0;
    // print ('Completed PostfixExpression case0');
  }

  /// <constructorInvocation> <selector>*
  void _case1(Lexer lexer) {
    constructorInvocation = ConstructorInvocation(lexer);

    selectors = [];
    int snapShot;
    try {
      while (true) {
        snapShot = lexer.currentIndex;
        selectors.add(Selector(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    caseNum = 1;
  }

  /// <primary> <selector>*
  void _case2(Lexer lexer) {
    primary = Primary(lexer);

    selectors = [];
    int snapShot;
    try {
      while (true) {
        snapShot = lexer.currentIndex;
        selectors.add(Selector(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    snapShot = lexer.currentIndex;
    PostfixOperator tmp1;
    AssignmentOperator tmp2;
    try {
      tmp1 = PostfixOperator(lexer);
      lexer.setPos(snapShot);
      tmp2 = AssignmentOperator(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    } finally {
      lexer.setPos(snapShot);
      if (tmp1 != null || tmp2 != null) {
        throw WrongTokenType('');
      }
    }

    caseNum = 2;
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return '$assignableExpression$posfixOperator';
        break;
      case 1:
        String rtn = '$constructorInvocation';
        selectors.forEach((s) => rtn += '$s');
        return rtn;
        break;
      case 2:
        String rtn = '$primary';
        selectors.forEach((s) => rtn += '$s');
        return rtn;
        break;
    }
    return '';
  }
}

/// <thisExpression> ::= this
class ThisExpression extends ParserType {
  ThisExpression(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'this');
    lexer.advance();
  }

  String toString() {
    return 'this';
  }
}

/// <newExpression> ::= new <typeNotVoid> ('.' <identifier>)? <arguments>
class NewExpression extends ParserType {
  TypeNotVoid typeNotVoid;
  Identifier identifier;
  Arguments arguments;

  NewExpression(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'new');
    lexer.advance();

    // print ('Started NewExpression');

    typeNotVoid = TypeNotVoid(lexer);

    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.PERIOD);
      lexer.advance();

      identifier = Identifier(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    arguments = Arguments(lexer);

    // print ('Finished NewExpression');
  }

  String toString() {
    return 'new $typeNotVoid' +
        (identifier != null ? '.$identifier' : '') +
        '$arguments';
  }
}

/// <constObjectExpression> ::= const <typeNotVoid> ('.' <identifier>)? <arguments>
class ConstObjectExpression extends ParserType {
  TypeNotVoid typeNotVoid;
  Identifier identifier;
  Arguments arguments;

  ConstObjectExpression(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'const');
    lexer.advance();

    typeNotVoid = TypeNotVoid(lexer);

    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.PERIOD);
      lexer.advance();

      identifier = Identifier(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    arguments = Arguments(lexer);
  }

  String toString() {
    return 'const $typeNotVoid' +
        (identifier != null ? '.$identifier' : '') +
        '$arguments';
  }
}

/// This differs from the official documentation (as of 23 Jun 2019).
/// See https://github.com/dart-lang/language/issues/381
///
/// <assignableExpression> ::= <primary> <assignableSelectorPart>+
/// | super <unconditionalAssignableSelector>
/// | <constructorInvocation> <assignableSelectorPart>+
/// | <identifier>
class AssignableExpression extends ParserType {
  int caseNum;
  Primary primary;
  List<AssignableSelectorPart> assignableSelectorParts;
  UnconditionalAssignableSelector unconditionalAssignableSelector;
  ConstructorInvocation constructorInvocation;
  Identifier identifier;

  AssignableExpression(Lexer lexer) {
    // print ('Starting AssignableExpression');
    // print (lexer.currentToken);
    // print (lexer.nextToken);

    assignableSelectorParts = [];
    int attempt, maxAttempts = 4;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            primary = Primary(lexer);

            assignableSelectorParts.add(AssignableSelectorPart(lexer));
            int snapShotInside;
            try {
              while (true) {
                snapShotInside = lexer.currentIndex;
                assignableSelectorParts.add(AssignableSelectorPart(lexer));
              }
            } on WrongTokenType {
              lexer.setPos(snapShotInside);
            }
            caseNum = 0;

            break;
          case 1:
            checkType(lexer, TType.KEYWORD);
            checkVal(lexer, 'super');
            lexer.advance();

            unconditionalAssignableSelector =
                UnconditionalAssignableSelector(lexer);

            caseNum = 1;
            break;
          case 2:
            identifier = Identifier(lexer);
            caseNum = 2;
            break;
          case 3:
            constructorInvocation = ConstructorInvocation(lexer);

            assignableSelectorParts.add(AssignableSelectorPart(lexer));
            int snapShotInside;
            try {
              while (true) {
                snapShotInside = lexer.currentIndex;
                assignableSelectorParts.add(AssignableSelectorPart(lexer));
              }
            } on WrongTokenType {
              lexer.setPos(snapShotInside);
            }

            // identifier = Identifier(lexer);

            caseNum = 3;
            break;
        }

        // print ('Completed AssignableExpression');
        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType(
          'Tried all options in AssignableExpression: ALL FAILED');
    }
  }

  String toString() {
    switch (caseNum) {
      case 0:
        String rtn = '$primary';
        assignableSelectorParts.forEach((p) => rtn += '$p');
        return rtn;
        break;
      case 1:
        return 'super $unconditionalAssignableSelector';
        break;
      case 2:
        return '$identifier';
        break;
      case 3:
        String rtn = '$constructorInvocation';
        assignableSelectorParts.forEach((p) => rtn += ' $p');
        return rtn; // + ' $identifier';
        break;
    }
    return '';
  }
}

/// <typeTest> ::= <isOperator> <typeNotVoid>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 143
class TypeTest extends ParserType {
  IsOperator isOperator;
  TypeNotVoid typeNotVoid;

  TypeTest(Lexer lexer) {
    isOperator = IsOperator(lexer);
    typeNotVoid = TypeNotVoid(lexer);
  }

  String toString() {
    return '$isOperator $typeNotVoid';
  }
}

/// <typeCast> ::= <asOperator> <typeNotVoid>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 143
class TypeCast extends ParserType {
  AsOperator asOperator;
  TypeNotVoid typeNotVoid;

  TypeCast(Lexer lexer) {
    asOperator = AsOperator(lexer);
    typeNotVoid = TypeNotVoid(lexer);
  }

  String toString() {
    return '$asOperator $typeNotVoid';
  }
}
