import '../ParserType.dart';
import '../ParserExceptions.dart';
import '../lexer/Lexer.dart';
import '../main/Expression.dart';
import '../main/MetaData.dart';
import '../core/Identifier.dart';
import '../core/Type.dart';
import '../sub_types/FormalParameterPart.dart';
import '../sub_types/Identifiers.dart';
import '../sub_types/Types.dart';

/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 21-22

/// <normalFormalParameters> ::= <normalFormalParameter> (',' <normalFormalParameter>)*
class NormalFormalParameters extends ParserType {
  List<NormalFormalParameter> normalFormalParameters;

  NormalFormalParameters(Lexer lexer) {
    normalFormalParameters = [];
    normalFormalParameters.add(NormalFormalParameter(lexer));

    int snapShot;
    try {
      while (isType(lexer, TType.COMMA)) {
        snapShot = lexer.currentIndex;
        lexer.advance();

        // These are for when there are both required and optional arguments
        // eg > void testFunc(bool someVar, {int a = 0}) {}
        checkNotType(lexer, TType.LEFT_BRACKET);
        checkNotType(lexer, TType.LEFT_CURLY);

        normalFormalParameters.add(NormalFormalParameter(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }
  }

  String toString() {
    String rtn = '';
    normalFormalParameters.forEach((p) => rtn += '$p, ');
    return rtn.substring(0, rtn.length - 2);
  }
}

/// <optionalFormalParameters> ::= <optionalPositionalFormalParameters> | <namedFormalParameters>
class OptionalFormalParameters extends ParserType {
  OptionalPositionalFormalParameters optionalPositionalFormalParameters;
  NamedFormalParameters namedFormalParameters;

  OptionalFormalParameters(Lexer lexer) {
    int snapShot = lexer.currentIndex;
    try {
      optionalPositionalFormalParameters =
          OptionalPositionalFormalParameters(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);

      namedFormalParameters = NamedFormalParameters(lexer);
    }
  }

  String toString() {
    return (optionalPositionalFormalParameters != null
        ? '$optionalPositionalFormalParameters'
        : '$namedFormalParameters');
  }
}

/// <optionalPositionalFormalParameters> ::=
/// '[' <defaultFormalParameter> (',' <defaultFormalParameter>)* ','? ']'
class OptionalPositionalFormalParameters extends ParserType {
  List<DefaultFormalParameter> defaultFormalParameters;
  bool hasComma;

  OptionalPositionalFormalParameters(Lexer lexer) {
    defaultFormalParameters = [];

    checkType(lexer, TType.LEFT_BRACKET);
    lexer.advance();

    defaultFormalParameters.add(DefaultFormalParameter(lexer));

    int snapShot;
    try {
      while (isType(lexer, TType.COMMA)) {
        snapShot = lexer.currentIndex;
        lexer.advance();

        defaultFormalParameters.add(DefaultFormalParameter(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    if (isType(lexer, TType.COMMA)) {
      hasComma = true;
      lexer.advance();
    } else {
      hasComma = false;
    }

    checkType(lexer, TType.RIGHT_BRACKET);
    lexer.advance();
  }

  String toString() {
    String rtn = '[';
    defaultFormalParameters.forEach((p) => rtn += '$p, ');
    return rtn.substring(0, rtn.length - 2) + (hasComma ? ',' : '') + ']';
  }
}

/// <namedFormalParameters> ::=
/// '{' <defaultNamedParameter> (',' <defaultNamedParameter>)* ','? '}'
class NamedFormalParameters extends ParserType {
  List<DefaultNamedParameter> defaultNamedParameters;
  bool hasComma;

  NamedFormalParameters(Lexer lexer) {
    defaultNamedParameters = [];

    checkType(lexer, TType.LEFT_CURLY);
    lexer.advance();

    defaultNamedParameters.add(DefaultNamedParameter(lexer));

    int snapShot;
    try {
      while (isType(lexer, TType.COMMA)) {
        snapShot = lexer.currentIndex;
        lexer.advance();

        defaultNamedParameters.add(DefaultNamedParameter(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    if (isType(lexer, TType.COMMA)) {
      hasComma = true;
      lexer.advance();
    } else {
      hasComma = false;
    }

    checkType(lexer, TType.RIGHT_CURLY);
    lexer.advance();
  }

  String toString() {
    String rtn = '{';
    defaultNamedParameters.forEach((p) => rtn += '$p, ');
    return rtn.substring(0, rtn.length - 2) + (hasComma ? ',' : '') + '}';
  }
}

/// <defaultFormalParameter> ::= <normalFormalParameter> ('=' <expression>)?
class DefaultFormalParameter extends ParserType {
  NormalFormalParameter normalFormalParameter;
  Expression expression;

  DefaultFormalParameter(Lexer lexer) {
    normalFormalParameter = NormalFormalParameter(lexer);

    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.EQUALS);
      lexer.advance();

      expression = Expression(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }
  }

  String toString() {
    return '$normalFormalParameter' +
        (expression != null ? '=$expression' : '');
  }
}

/// <defaultNamedParameter> ::= <normalFormalParameter> ('=' <expression>)?
/// | <normalFormalParameter> ( ':' <expression>)?
class DefaultNamedParameter extends ParserType {
  NormalFormalParameter normalFormalParameter;
  Expression expression;
  bool isEquals;

  DefaultNamedParameter(Lexer lexer) {
    normalFormalParameter = NormalFormalParameter(lexer);

    int snapShot = lexer.currentIndex;
    try {
      if (isType(lexer, TType.EQUALS)) {
        isEquals = true;
      } else if (isType(lexer, TType.COLON)) {
        isEquals = false;
      } else {
        checkType(lexer, TType.EQUALS);
      }
      lexer.advance();

      expression = Expression(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }
  }

  String toString() {
    return '$normalFormalParameter' +
        (expression != null ? (isEquals ? '=' : ':') + '$expression' : '');
  }
}

/// <normalFormalParameter> ::= <functionFormalParameter>
/// | <fieldFormalParameter>
/// | <simpleFormalParameter>
class NormalFormalParameter extends ParserType {
  int caseNum;
  FunctionFormalParameter functionFormalParameter;
  FieldFormalParameter fieldFormalParameter;
  SimpleFormalParameter simpleFormalParameter;

  NormalFormalParameter(Lexer lexer) {
    int attempt, maxAttempts = 3;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            functionFormalParameter = FunctionFormalParameter(lexer);
            caseNum = 0;
            break;
          case 1:
            fieldFormalParameter = FieldFormalParameter(lexer);
            caseNum = 1;
            break;
          case 2:
            simpleFormalParameter = SimpleFormalParameter(lexer);
            caseNum = 2;
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType(
          'Tried all options in NormalFormalParameter: ALL FAILED');
    }
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return '$functionFormalParameter';
        break;
      case 1:
        return '$fieldFormalParameter';
        break;
      case 2:
        return '$simpleFormalParameter';
        break;
    }
    return '';
  }
}

/// <functionFormalParameter> ::=
/// <metadata> covariant? <type>? <identifier> <formalParameterPart>
class FunctionFormalParameter extends ParserType {
  MetaData metaData;
  Type type;
  Identifier identifier;
  FormalParameterPart formalParameterPart;
  bool isCovariant;

  FunctionFormalParameter(Lexer lexer) {
    metaData = MetaData(lexer);

    if (isType(lexer, TType.KEYWORD) && isVal(lexer, 'covariant')) {
      isCovariant = true;
      lexer.advance();
    } else {
      isCovariant = false;
    }

    int snapShot = lexer.currentIndex;
    try {
      type = Type(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    identifier = Identifier(lexer);
    formalParameterPart = FormalParameterPart(lexer);
  }

  String toString() {
    return '$metaData' +
        (isCovariant ? 'covariant ' : '') +
        (type != null ? '$type ' : '') +
        '$identifier $formalParameterPart';
  }
}

/// <fieldFormalParameter> ::=
/// <metadata> <finalConstVarOrType>? this '.' <identifier> <formalParameterPart>?
class FieldFormalParameter extends ParserType {
  MetaData metaData;
  FinalConstVarOrType finalConstVarOrType;
  Identifier identifier;
  FormalParameterPart formalParameterPart;

  FieldFormalParameter(Lexer lexer) {
    metaData = MetaData(lexer);

    int snapShot = lexer.currentIndex;
    try {
      finalConstVarOrType = FinalConstVarOrType(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'this');
    lexer.advance();

    checkType(lexer, TType.PERIOD);
    lexer.advance();

    identifier = Identifier(lexer);

    snapShot = lexer.currentIndex;
    try {
      formalParameterPart = FormalParameterPart(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }
  }

  String toString() {
    return '$metaData' +
        (finalConstVarOrType != null ? '$finalConstVarOrType ' : '') +
        'this.$identifier' +
        (formalParameterPart != null ? ' $formalParameterPart' : '');
  }
}

/// <simpleFormalParameter> ::= <declaredIdentifier>
/// | <metadata> covariant? <identifier>
class SimpleFormalParameter extends ParserType {
  DeclaredIdentifier declaredIdentifier;
  MetaData metaData;
  Identifier identifier;
  bool isCovariant;

  SimpleFormalParameter(Lexer lexer) {
    int snapShot = lexer.currentIndex;
    try {
      declaredIdentifier = DeclaredIdentifier(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);

      metaData = MetaData(lexer);

      if (isType(lexer, TType.KEYWORD) && isVal(lexer, 'covariant')) {
        isCovariant = true;
        lexer.advance();
      } else {
        isCovariant = false;
      }

      identifier = Identifier(lexer);
    }
  }

  String toString() {
    if (declaredIdentifier != null) {
      return '$declaredIdentifier';
    }
    return '$metaData' + (isCovariant ? 'covariant ' : '') + '$identifier';
  }
}
