import '../ParserType.dart';
import '../ParserExceptions.dart';
import '../lexer/Lexer.dart';
import '../main/MetaData.dart';
import '../main/Expression.dart';
import '../sub_types/Types.dart';
import '../core/Identifier.dart';

/// <declaredIdentifier> ::= <metadata> covariant? <finalConstVarOrType> <identifier>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 14
class DeclaredIdentifier extends ParserType {
  MetaData metaData;
  FinalConstVarOrType finalConstVarOrType;
  Identifier identifier;
  bool isCovariant;

  DeclaredIdentifier(Lexer lexer) {
    metaData = MetaData(lexer);

    if (isType(lexer, TType.KEYWORD) && isVal(lexer, 'covariant')) {
      isCovariant = true;
      lexer.advance();
    } else {
      isCovariant = false;
    }

    finalConstVarOrType = FinalConstVarOrType(lexer);
    identifier = Identifier(lexer);
  }

  String toString() {
    return '$metaData' +
        (isCovariant ? 'covariant ' : '') +
        '$finalConstVarOrType ' +
        '$identifier';
  }
}

/// <initializedIdentifier> ::= <identifier> ('=' <expression>)?
class InitializedIdentifier extends ParserType {
  Identifier identifier;
  Expression expression;

  InitializedIdentifier(Lexer lexer) {
    identifier = Identifier(lexer);

    int snapShot = lexer.currentIndex;
    try {
      if (isType(lexer, TType.EQUALS)) {
        lexer.advance();
        expression = Expression(lexer);
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }
  }

  String toString() {
    return '$identifier' + (expression != null ? ' = $expression' : '');
  }
}

/// <typeIdentifier> denotes an identifier which can
/// be the name of a type, that is, it denotes an <IDENTIFIER> which is not a
/// <BUILT_IN_IDENTIFIER>.
///
/// <BUILT_IN_IDENTIFIER> ::= abstract
/// | as
/// | covariant
/// | deferred
/// | dynamic
/// | export
/// | external
/// | factory
/// | Function
/// | get
/// | implements
/// | import
/// | interface
/// | library
/// | operator
/// | mixin
/// | part
/// | set
/// | static
/// | typedef
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 176 and 141
class TypeIdentifier extends ParserType {
  String val;

  TypeIdentifier(Lexer lexer) {
    checkNotVal(lexer, 'abstract');
    checkNotVal(lexer, 'as');
    checkNotVal(lexer, 'covariant');
    checkNotVal(lexer, 'deferred');
    // checkNotVal(lexer, 'dynamic');
    checkNotVal(lexer, 'export');
    checkNotVal(lexer, 'external');
    checkNotVal(lexer, 'factory');
    checkNotVal(lexer, 'Function');
    checkNotVal(lexer, 'get');
    checkNotVal(lexer, 'implements');
    checkNotVal(lexer, 'import');
    checkNotVal(lexer, 'interface');
    checkNotVal(lexer, 'library');
    checkNotVal(lexer, 'operator');
    checkNotVal(lexer, 'mixin');
    checkNotVal(lexer, 'part');
    checkNotVal(lexer, 'set');
    checkNotVal(lexer, 'static');
    checkNotVal(lexer, 'typedef');

    checkNotVal(lexer, 'this');

    if (!(isType(lexer, TType.IDENTIFIER) || isType(lexer, TType.KEYWORD))) {
      throw WrongTokenType(
          'Expected ID or KEYWORD, got ${lexer.currentToken} in TypeIdentifier');
    }

    val = lexer.currentToken.val;
    lexer.advance();
  }

  String toString() {
    return '$val';
  }
}
