import '../ParserType.dart';
import '../ParserExceptions.dart';
import '../lexer/Lexer.dart';
import '../lists/Arguments.dart';
import '../core/Identifier.dart';
import '../sub_types/Assertion.dart';
import '../sub_types/Expressions.dart';
import '../sub_types/CascadeSection.dart';

/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 40

/// <initializerListEntry> ::= super <arguments>
/// | super '.' <identifier> <arguments>
/// | <fieldInitializer>
/// | <assertion>
class InitializerListEntry extends ParserType {
  int caseNum;
  Arguments arguments;
  Identifier identifier;
  FieldInitializer fieldInitializer;
  Assertion assertion;

  InitializerListEntry(Lexer lexer) {
    int attempt, maxAttempts = 4;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
          case 2:
            _case2(lexer);
            break;
          case 3:
            _case3(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType('Tried all options in SuperClass: ALL FAILED');
    }
  }

  /// super <arguments>
  void _case0(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'super');
    lexer.advance();

    arguments = Arguments(lexer);

    caseNum = 0;
  }

  /// super '.' <identifier> <arguments>
  void _case1(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'super');
    lexer.advance();

    checkType(lexer, TType.PERIOD);
    lexer.advance();

    identifier = Identifier(lexer);
    arguments = Arguments(lexer);

    caseNum = 1;
  }

  /// <fieldInitializer>
  void _case2(Lexer lexer) {
    fieldInitializer = FieldInitializer(lexer);
    caseNum = 2;
  }

  /// <assertion>
  void _case3(Lexer lexer) {
    assertion = Assertion(lexer);
    caseNum = 3;
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return 'super $arguments';
        break;
      case 1:
        return 'super.$identifier $arguments';
        break;
      case 2:
        return '$fieldInitializer';
        break;
      case 3:
        return '$assertion';
        break;
    }
    return '';
  }
}

/// <fieldInitializer> ::=
/// (this '.')? <identifier> '=' <conditionalExpression> <cascadeSection>*
class FieldInitializer extends ParserType {
  Identifier identifier;
  ConditionalExpression conditionalExpression;
  List<CascadeSection> cascadeSections;
  bool isThis;

  FieldInitializer(Lexer lexer) {
    isThis = false;
    cascadeSections = [];

    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.KEYWORD);
      checkVal(lexer, 'this');
      lexer.advance();

      checkType(lexer, TType.PERIOD);
      lexer.advance();

      isThis = true;
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    identifier = Identifier(lexer);

    checkType(lexer, TType.EQUALS);
    lexer.advance();

    conditionalExpression = ConditionalExpression(lexer);

    try {
      while (true) {
        snapShot = lexer.currentIndex;
        cascadeSections.add(CascadeSection(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }
  }

  String toString() {
    String rtn = (isThis ? 'this.' : '') + '$identifier=$conditionalExpression';
    cascadeSections.forEach((s) => rtn += ' $s');
    return rtn;
  }
}
