import '../ParserType.dart';
import '../ParserExceptions.dart';
import '../lexer/Lexer.dart';
import '../core/Identifier.dart';
import '../core/Operator.dart';
import '../main/Expression.dart';
import '../lists/Elements.dart';
import 'Types.dart';

/// <nullLiteral> ::= null
class NullLiteral extends ParserType {
  NullLiteral(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'null');
    lexer.advance();
  }

  String toString() {
    return 'null';
  }
}

/// <booleanLiteral> ::= true | false
class BooleanLiteral extends ParserType {
  String val;

  BooleanLiteral(Lexer lexer) {
    checkType(lexer, TType.BOOLEAN);
    val = lexer.currentToken.val;
    lexer.advance();
  }

  String toString() {
    return '$val';
  }
}

/// <numericLiteral> ::= <NUMBER> | <HEX_NUMBER>
class NumericLiteral extends ParserType {
  String val;

  NumericLiteral(Lexer lexer) {
    try {
      checkType(lexer, TType.NUMBER);
      val = lexer.currentToken.val;
    } on WrongTokenType {
      checkType(lexer, TType.HEX_NUMBER);
      val = lexer.currentToken.val;
    }
    lexer.advance();
  }

  String toString() {
    return '$val';
  }
}

/// <stringLiteral> ::= (<multilineString> | <singleLineString>)+
class StringLiteral extends ParserType {
  String val;

  StringLiteral(Lexer lexer) {
    checkType(lexer, TType.STRING);
    val = lexer.currentToken.val;
    lexer.advance();
  }

  String toString() {
    return '$val';
  }
}

/// <symbolLiteral> ::= '#' (<operator> | (<identifier> ('.' <identifier>)*))
class SymbolLiteral extends ParserType {
  int caseNum;
  Operator operator_;
  List<Identifier> identifiers;

  SymbolLiteral(Lexer lexer) {
    checkType(lexer, TType.HASH);
    lexer.advance();

    identifiers = [];
    int snapShot = lexer.currentIndex;
    try {
      operator_ = Operator(lexer);
      caseNum = 0;
    } on WrongTokenType {
      lexer.setPos(snapShot);

      identifiers.add(Identifier(lexer));
      while (isType(lexer, TType.PERIOD)) {
        snapShot = lexer.currentIndex;
        lexer.advance();
        identifiers.add(Identifier(lexer));
      }

      caseNum = 1;
    }
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return '#$operator_';
        break;
      case 1:
        String rtn = '';
        identifiers.forEach((i) => rtn += '.$i');
        return '#' + rtn.substring(1);
        break;
    }
    return '';
  }
}

/// <mapLiteral> ::= const? <typeArguments>? '{' <mapLiteralEntry> (',' <mapLiteralEntry>)* ','? '}'
class MapLiteral extends ParserType {
  TypeArguments typeArguments;
  List<MapLiteralEntry> mapLiteralEntries;
  bool isConst;
  bool hasComma;

  MapLiteral(Lexer lexer) {
    mapLiteralEntries = [];

    if (isType(lexer, TType.KEYWORD) && isVal(lexer, 'const')) {
      isConst = true;
      lexer.advance();
    } else {
      isConst = false;
    }

    int snapShot = lexer.currentIndex;
    try {
      typeArguments = TypeArguments(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    checkType(lexer, TType.LEFT_CURLY);
    lexer.advance();

    mapLiteralEntries.add(MapLiteralEntry(lexer));

    try {
      while (isType(lexer, TType.COMMA)) {
        snapShot = lexer.currentIndex;
        lexer.advance();
        mapLiteralEntries.add(MapLiteralEntry(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    if (isType(lexer, TType.COMMA)) {
      lexer.advance();
      hasComma = true;
    } else {
      hasComma = false;
    }

    checkType(lexer, TType.RIGHT_CURLY);
    lexer.advance();
  }

  String toString() {
    String rtn = '';
    mapLiteralEntries.forEach((e) => rtn += ', $e');
    return (isConst ? 'const ' : '') +
        (typeArguments != null ? '$typeArguments' : '') +
        '{' +
        rtn.substring(2) +
        (hasComma ? ',' : '') +
        '}';
  }
}

/// <mapLiteralEntry> ::= <expression> ':' <expression>
class MapLiteralEntry extends ParserType {
  Expression expression1;
  Expression expression2;

  MapLiteralEntry(Lexer lexer) {
    expression1 = Expression(lexer);

    checkType(lexer, TType.COLON);
    lexer.advance();

    expression2 = Expression(lexer);
  }

  String toString() {
    return '$expression1:$expression2';
  }
}

/// <setLiteral> ::= const? <typeArguments>? '{' <expression> (',' <expression>)* ','? '}'
class SetLiteral extends ParserType {
  TypeArguments typeArguments;
  List<Expression> expressions;
  bool isConst;
  bool hasComma;

  SetLiteral(Lexer lexer) {
    expressions = [];

    if (isType(lexer, TType.KEYWORD) && isVal(lexer, 'const')) {
      isConst = true;
      lexer.advance();
    } else {
      isConst = false;
    }

    int snapShot = lexer.currentIndex;
    try {
      typeArguments = TypeArguments(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    checkType(lexer, TType.LEFT_CURLY);
    lexer.advance();

    expressions.add(Expression(lexer));

    try {
      while (isType(lexer, TType.COMMA)) {
        snapShot = lexer.currentIndex;
        lexer.advance();
        expressions.add(Expression(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    if (isType(lexer, TType.COMMA)) {
      lexer.advance();
      hasComma = true;
    } else {
      hasComma = false;
    }

    checkType(lexer, TType.RIGHT_CURLY);
    lexer.advance();
  }

  String toString() {
    String rtn = '';
    expressions.forEach((e) => rtn += ', $e');
    return (isConst ? 'const ' : '') +
        (typeArguments != null ? '$typeArguments' : '') +
        '{' +
        rtn.substring(2) +
        (hasComma ? ',' : '') +
        '}';
  }
}

/// <setOrMapLiteral> ::= const? <typeArguments>? '{' <elements>? '}'
class SetOrMapLiteral extends ParserType {
  TypeArguments typeArguments;
  Elements elements;
  bool isConst;

  SetOrMapLiteral(Lexer lexer) {
    if (isType(lexer, TType.KEYWORD) && isVal(lexer, 'const')) {
      isConst = true;
      lexer.advance();
    } else {
      isConst = false;
    }

    int snapShot = lexer.currentIndex;
    try {
      typeArguments = TypeArguments(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    checkType(lexer, TType.LEFT_CURLY);
    lexer.advance();

    snapShot = lexer.currentIndex;
    try {
      elements = Elements(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    checkType(lexer, TType.RIGHT_CURLY);
    lexer.advance();
  }

  String toString() {
    return (isConst ? 'const ' : '') +
        (typeArguments != null ? '$typeArguments' : '') +
        '{' +
        (elements != null ? '$elements' : '') +
        '}';
  }
}

/// <listLiteral> ::= const? <typeArguments>? '[' <elements>? ']'
class ListLiteral extends ParserType {
  TypeArguments typeArguments;
  Elements elements;
  bool isConst;

  ListLiteral(Lexer lexer) {
    if (isType(lexer, TType.KEYWORD) && isVal(lexer, 'const')) {
      isConst = true;
      lexer.advance();
    } else {
      isConst = false;
    }

    int snapShot = lexer.currentIndex;
    try {
      typeArguments = TypeArguments(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    checkType(lexer, TType.LEFT_BRACKET);
    lexer.advance();

    // print ('started ListLiteral');

    snapShot = lexer.currentIndex;
    try {
      elements = Elements(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    checkType(lexer, TType.RIGHT_BRACKET);
    lexer.advance();

    // print ('finished ListLiteral');
  }

  String toString() {
    return (isConst ? 'const ' : '') +
        (typeArguments != null ? '$typeArguments' : '') +
        '[' +
        (elements != null ? '$elements' : '') +
        ']';
  }
}
