import '../ParserType.dart';
import '../ParserExceptions.dart';
import '../lexer/Lexer.dart';

/// <assignmentOperator> ::= '=' | <compoundAssignmentOperator>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 125
class AssignmentOperator extends ParserType {
  String val;
  CompoundAssignmentOperator compoundAssignmentOperator;

  AssignmentOperator(Lexer lexer) {
    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.EQUALS);
      lexer.advance();
      val = '=';
    } on WrongTokenType {
      lexer.setPos(snapShot);

      compoundAssignmentOperator = CompoundAssignmentOperator(lexer);
    }
  }

  String toString() {
    return (compoundAssignmentOperator != null
        ? '$compoundAssignmentOperator'
        : '$val');
  }
}

/// <compoundAssignmentOperator> ::= '*='
/// | '/='
/// | '~/='
/// | '%='
/// | '+='
/// | '-='
/// | '<<='
/// | '>>='
/// | '>>>='
/// | '&='
/// | '^='
/// | '|='
/// | '??='
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 130
class CompoundAssignmentOperator extends ParserType {
  String val;

  CompoundAssignmentOperator(Lexer lexer) {
    if (isType(lexer, TType.MULTIPLY)) {
      _checkEquals(lexer);
      val = '*=';
    } else if (isType(lexer, TType.DIVIDE)) {
      _checkEquals(lexer);
      val = '/=';
    } else if (isType(lexer, TType.PERCENT)) {
      _checkEquals(lexer);
      val = '%=';
    } else if (isType(lexer, TType.PLUS)) {
      _checkEquals(lexer);
      val = '+=';
    } else if (isType(lexer, TType.MINUS)) {
      _checkEquals(lexer);
      val = '-=';
    } else if (isType(lexer, TType.AMPERSAND)) {
      _checkEquals(lexer);
      val = '&=';
    } else if (isType(lexer, TType.CARAT)) {
      _checkEquals(lexer);
      val = '^=';
    } else if (isType(lexer, TType.PIPE)) {
      _checkEquals(lexer);
      val = '|=';
    } else if (isType(lexer, TType.TILDE)) {
      lexer.advance();
      checkType(lexer, TType.DIVIDE);
      _checkEquals(lexer);
      val = '~/=';
    } else if (isType(lexer, TType.Q_MARK)) {
      lexer.advance();
      checkType(lexer, TType.Q_MARK);
      _checkEquals(lexer);
      val = '??=';
    } else if (isType(lexer, TType.LEFT_ANGLE)) {
      lexer.advance();
      checkType(lexer, TType.LEFT_ANGLE);
      _checkEquals(lexer);
      val = '<<=';
    } else if (isType(lexer, TType.RIGHT_ANGLE)) {
      lexer.advance();
      checkType(lexer, TType.RIGHT_ANGLE);
      lexer.advance();
      if (isType(lexer, TType.RIGHT_ANGLE)) {
        val = '>>>=';
        lexer.advance();
      } else {
        val = '>>=';
      }
      _checkEquals(lexer, false);
    } else {
      throw WrongTokenType('Not a CompoundAssignmentOperator');
    }
  }

  void _checkEquals(Lexer lexer, [bool advance = true]) {
    (advance ? lexer.advance() : '');
    checkType(lexer, TType.EQUALS);
    lexer.advance();
  }

  String toString() {
    return '$val';
  }
}

/// <relationalOperator> ::= '>=' | '>' | '<=' | '<'
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 133-136
class RelationalOperator extends ParserType {
  String val;

  RelationalOperator(Lexer lexer) {
    if (isType(lexer, TType.RIGHT_ANGLE)) {
      lexer.advance();
      if (isType(lexer, TType.EQUALS)) {
        val = '>=';
        lexer.advance();
      } else {
        val = '>';
      }
    } else if (isType(lexer, TType.LEFT_ANGLE)) {
      lexer.advance();
      if (isType(lexer, TType.EQUALS)) {
        val = '<=';
        lexer.advance();
      } else {
        val = '<';
      }
    } else {
      throw WrongTokenType('Failed to get RelationalOperator');
    }
  }

  String toString() {
    return '$val';
  }
}

/// <bitwiseOperator> ::= '&' | '^' | '|'
class BitwiseOperator extends ParserType {
  String val;

  BitwiseOperator(Lexer lexer) {
    if (isType(lexer, TType.AMPERSAND)) {
      val = '&';
    } else if (isType(lexer, TType.CARAT)) {
      val = '^';
    } else if (isType(lexer, TType.PIPE)) {
      val = '|';
    } else {
      throw WrongTokenType('Failed to get BitwiseOperator');
    }
    lexer.advance();
  }

  String toString() {
    return '$val';
  }
}

/// <shiftOperator> ::= '<<' | '>>' | '>>>'
class ShiftOperator extends ParserType {
  String val;

  ShiftOperator(Lexer lexer) {
    if (isType(lexer, TType.LEFT_ANGLE)) {
      lexer.advance();
      checkType(lexer, TType.LEFT_ANGLE);
      val = '<<';
    } else if (isType(lexer, TType.RIGHT_ANGLE)) {
      lexer.advance();
      checkType(lexer, TType.RIGHT_ANGLE);
      lexer.advance();
      if (isType(lexer, TType.RIGHT_ANGLE)) {
        val = '>>>';
      } else {
        val = '>>';
        lexer.advance(-1);
      }
    } else {
      throw WrongTokenType('Failed to get ShiftOperator');
    }
    lexer.advance();
  }

  String toString() {
    return '$val';
  }
}

/// <additiveOperator> ::= '+' | '-'
class AdditiveOperator extends ParserType {
  String val;

  AdditiveOperator(Lexer lexer) {
    if (isType(lexer, TType.PLUS)) {
      lexer.advance();
      val = '+';
    } else if (isType(lexer, TType.MINUS)) {
      lexer.advance();
      val = '-';
    } else {
      throw WrongTokenType('Failed to get AdditiveOperator');
    }
  }

  String toString() {
    return '$val';
  }
}

/// <multiplicativeOperator> ::= '*' | '/' | '%' | '~/'
class MultiplicativeOperator extends ParserType {
  String val;

  MultiplicativeOperator(Lexer lexer) {
    if (isType(lexer, TType.MULTIPLY)) {
      val = '*';
    } else if (isType(lexer, TType.DIVIDE)) {
      val = '/';
    } else if (isType(lexer, TType.PERCENT)) {
      val = '%';
    } else if (isType(lexer, TType.TILDE)) {
      lexer.advance();
      checkType(lexer, TType.DIVIDE);
      val = '~/';
    } else {
      throw WrongTokenType('Failed to get MultiplicativeOperator');
    }
    lexer.advance();
  }

  String toString() {
    return '$val';
  }
}

/// <prefixOperator> ::= <minusOperator> | <negationOperator> | <tildeOperator>
class PrefixOperator extends ParserType {
  int caseNum;
  MinusOperator minusOperator;
  NegationOperator negationOperator;
  TildeOperator tildeOperator;

  PrefixOperator(Lexer lexer) {
    int attempt, maxAttempts = 3;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
          case 2:
            _case2(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType('Tried all options in PrefixOperator: ALL FAILED');
    }
  }

  /// <minusOperator>
  void _case0(Lexer lexer) {
    minusOperator = MinusOperator(lexer);
    caseNum = 0;
  }

  /// <negationOperator>
  void _case1(Lexer lexer) {
    negationOperator = NegationOperator(lexer);
    caseNum = 1;
  }

  /// <tildeOperator>
  void _case2(Lexer lexer) {
    tildeOperator = TildeOperator(lexer);
    caseNum = 2;
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return '$minusOperator';
        break;
      case 1:
        return '$negationOperator';
        break;
      case 2:
        return '$tildeOperator';
        break;
    }
    return '';
  }
}

/// <minusOperator> ::= '-'
class MinusOperator extends ParserType {
  String val;

  MinusOperator(Lexer lexer) {
    checkType(lexer, TType.MINUS);
    lexer.advance();
    checkNotType(lexer, TType.MINUS); // Necessary to catch increment Operator
    val = '-';
  }

  String toString() {
    return '$val';
  }
}

/// <negationOperator> ::= '!'
class NegationOperator extends ParserType {
  String val;

  NegationOperator(Lexer lexer) {
    checkType(lexer, TType.EXCLAMATION);
    lexer.advance();
    val = '!';
  }

  String toString() {
    return '$val';
  }
}

/// <tildeOperator> ::= '~'
class TildeOperator extends ParserType {
  String val;

  TildeOperator(Lexer lexer) {
    checkType(lexer, TType.TILDE);
    lexer.advance();
    val = '~';
  }

  String toString() {
    return '$val';
  }
}

/// <postfixOperator> ::= <incrementOperator>
class PostfixOperator extends ParserType {
  IncrementOperator incrementOperator;

  PostfixOperator(Lexer lexer) {
    incrementOperator = IncrementOperator(lexer);
  }

  String toString() {
    return '$incrementOperator';
  }
}

/// <equalityOperator> ::= '==' | '!='
class EqualityOperator extends ParserType {
  String val;

  EqualityOperator(Lexer lexer) {
    if (isType(lexer, TType.EQUALS)) {
      lexer.advance();
      checkType(lexer, TType.EQUALS);
      lexer.advance();
      val = '==';
    } else if (isType(lexer, TType.EXCLAMATION)) {
      lexer.advance();
      checkType(lexer, TType.EQUALS);
      lexer.advance();
      val = '!=';
    } else {
      throw WrongTokenType('Expected "=" or "!" and got ${lexer.currentToken}');
    }
  }

  String toString() {
    return '$val';
  }
}

/// <incrementOperator> ::= '++' | '--'
class IncrementOperator extends ParserType {
  String val;

  IncrementOperator(Lexer lexer) {
    if (isType(lexer, TType.PLUS)) {
      lexer.advance();
      checkType(lexer, TType.PLUS);
      val = '++';
    } else if (isType(lexer, TType.MINUS)) {
      lexer.advance();
      checkType(lexer, TType.MINUS);
      val = '--';
    } else {
      // print ('FAILED ---------------------------------------');
      throw WrongTokenType('Failed to get IncrementOperator');
    }
    lexer.advance();
  }

  String toString() {
    return '$val';
  }
}

/// <isOperator> ::= is '!'?
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 143
class IsOperator extends ParserType {
  bool hasExclamation;

  IsOperator(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'is');
    lexer.advance();

    if (isType(lexer, TType.EXCLAMATION)) {
      hasExclamation = true;
      lexer.advance();
    } else {
      hasExclamation = false;
    }
  }

  String toString() {
    return 'is' + (hasExclamation ? '!' : '');
  }
}

/// <asOperator> ::= as
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 144
class AsOperator extends ParserType {
  AsOperator(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'as');
    lexer.advance();
  }

  String toString() {
    return 'as';
  }
}
