import '../ParserType.dart';
import '../ParserExceptions.dart';
import '../lexer/Lexer.dart';
import '../core/Identifier.dart';
import '../core/Type.dart';

/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 176-177

/// <normalParameterTypes> ::= <normalParameterType> (',' <normalParameterType>)*
class NormalParameterTypes extends ParserType {
  List<NormalParameterType> normalParameterTypes;

  NormalParameterTypes(Lexer lexer) {
    normalParameterTypes = [];
    normalParameterTypes.add(NormalParameterType(lexer));

    int snapShot;
    try {
      while (isType(lexer, TType.COMMA)) {
        snapShot = lexer.currentIndex;
        lexer.advance();

        normalParameterTypes.add(NormalParameterType(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }
  }
}

/// <normalParameterType> ::= <typedIdentifier> | <type>
class NormalParameterType extends ParserType {
  TypedIdentifier typedIdentifier;
  Type type;

  NormalParameterType(Lexer lexer) {
    int snapShot = lexer.currentIndex;
    try {
      typedIdentifier = TypedIdentifier(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);

      type = Type(lexer);
    }
  }
}

/// <optionalParameterTypes> ::= <optionalPositionalParameterTypes> | <namedParameterTypes>
class OptionalParamterTypes extends ParserType {
  OptionalPositionalParameterTypes optionalPositionalParameterTypes;
  NamedParameterTypes namedParameterTypes;

  OptionalParamterTypes(Lexer lexer) {
    int snapShot = lexer.currentIndex;
    try {
      optionalPositionalParameterTypes =
          OptionalPositionalParameterTypes(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);

      namedParameterTypes = NamedParameterTypes(lexer);
    }
  }
}

/// <optionalPositionalParameterTypes> ::= '[' <normalParameterTypes> ','? ']'
class OptionalPositionalParameterTypes extends ParserType {
  NormalParameterTypes normalParameterTypes;
  bool hasComma;

  OptionalPositionalParameterTypes(Lexer lexer) {
    checkType(lexer, TType.LEFT_BRACKET);
    lexer.advance();

    normalParameterTypes = NormalParameterTypes(lexer);

    if (isType(lexer, TType.COMMA)) {
      hasComma = true;
      lexer.advance();
    } else {
      hasComma = false;
    }

    checkType(lexer, TType.RIGHT_BRACKET);
    lexer.advance();
  }
}

/// <namedParameterTypes> ::= '{' <typedIdentifier> (',' <typedIdentifier>)* ','? '}'
class NamedParameterTypes extends ParserType {
  List<TypedIdentifier> typedIdentifiers;
  bool hasComma;

  NamedParameterTypes(Lexer lexer) {
    typedIdentifiers = [];

    checkType(lexer, TType.LEFT_CURLY);
    lexer.advance();

    typedIdentifiers.add(TypedIdentifier(lexer));

    int snapShot;
    try {
      while (isType(lexer, TType.COMMA)) {
        snapShot = lexer.currentIndex;
        lexer.advance();

        typedIdentifiers.add(TypedIdentifier(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    if (isType(lexer, TType.COMMA)) {
      hasComma = true;
      lexer.advance();
    } else {
      hasComma = false;
    }

    checkType(lexer, TType.RIGHT_CURLY);
    lexer.advance();
  }
}

/// <typedIdentifier> ::= <type> <identifier>
class TypedIdentifier extends ParserType {
  Type type;
  Identifier identifier;

  TypedIdentifier(Lexer lexer) {
    type = Type(lexer);
    identifier = Identifier(lexer);
  }
}
