import '../ParserExceptions.dart';
import '../ParserType.dart';
import '../lexer/Lexer.dart';
import '../lists/FormalParameterList.dart';
import '../lists/Initializers.dart';
import '../lists/Arguments.dart';
import '../core/Identifier.dart';
import '../core/Qualified.dart';
import '../core/Operator.dart';
import '../core/Type.dart';
import '../main/MetaData.dart';
import '../sub_types/FormalParameterPart.dart';

/// <methodSignature> ::= <constructorSignature> <initializers>?
/// | <factoryConstructorSignature>
/// | static? <functionSignature>
/// | static? <getterSignature>
/// | static? <setterSignature>
/// | <operatorSignature>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 26
class MethodSignature extends ParserType {
  int caseNum;
  ConstructorSignature constructorSignature;
  Initializers initializers;
  FactoryConstructorSignature factoryConstructorSignature;
  FunctionSignature functionSignature;
  GetterSignature getterSignature;
  SetterSignature setterSignature;
  OperatorSignature operatorSignature;
  bool isStatic;

  MethodSignature(Lexer lexer) {
    isStatic = false;

    int attempt, maxAttempts = 6;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
          case 2:
            _case2(lexer);
            break;
          case 3:
            _case3(lexer);
            break;
          case 4:
            _case4(lexer);
            break;
          case 5:
            _case5(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType('Tried all options in MethodSignature: ALL FAILED');
    }
  }

  /// <constructorSignature> <initializers>?
  void _case0(Lexer lexer) {
    constructorSignature = ConstructorSignature(lexer);

    int snapShot = lexer.currentIndex;
    try {
      initializers = Initializers(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    caseNum = 0;
  }

  /// <factoryConstructorSignature>
  void _case1(Lexer lexer) {
    factoryConstructorSignature = FactoryConstructorSignature(lexer);
    caseNum = 1;
  }

  /// static? <functionSignature>
  void _case2(Lexer lexer) {
    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.KEYWORD);
      checkVal(lexer, 'static');
      lexer.advance();

      isStatic = true;
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    functionSignature = FunctionSignature(lexer);

    caseNum = 2;
  }

  /// static? <getterSignature>
  void _case3(Lexer lexer) {
    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.KEYWORD);
      checkVal(lexer, 'static');
      lexer.advance();

      isStatic = true;
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    getterSignature = GetterSignature(lexer);

    caseNum = 3;
  }

  /// static? <setterSignature>
  void _case4(Lexer lexer) {
    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.KEYWORD);
      checkVal(lexer, 'static');
      lexer.advance();

      isStatic = true;
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    setterSignature = SetterSignature(lexer);

    caseNum = 4;
  }

  /// <operatorSignature>
  void _case5(Lexer lexer) {
    operatorSignature = OperatorSignature(lexer);
    caseNum = 5;
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return '$constructorSignature' +
            (initializers != null ? ' $initializers' : '');
        break;
      case 1:
        return '$factoryConstructorSignature';
        break;
      case 2:
        return (isStatic ? 'static ' : '') + '$functionSignature';
        break;
      case 3:
        return (isStatic ? 'static ' : '') + '$getterSignature';
        break;
      case 4:
        return (isStatic ? 'static ' : '') + '$setterSignature';
        break;
      case 5:
        return '$operatorSignature';
        break;
    }
    return '';
  }
}

/// <constructorSignature> ::=
/// <identifier> ('.' <identifier>)? <formalParameterList>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 37
class ConstructorSignature extends ParserType {
  Identifier identifier1;
  Identifier identifier2;
  FormalParameterList formalParameterList;

  ConstructorSignature(Lexer lexer) {
    identifier1 = Identifier(lexer);

    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.PERIOD);
      lexer.advance();

      identifier2 = Identifier(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    formalParameterList = FormalParameterList(lexer);
  }

  String toString() {
    return '$identifier1' +
        (identifier2 != null ? '.$identifier2' : '') +
        '$formalParameterList';
  }
}

/// <factoryConstructorSignature> ::=
/// factory <identifier> ('.' <identifier>)? <formalParameterList>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 43
class FactoryConstructorSignature extends ParserType {
  Identifier identifier1;
  Identifier identifier2;
  FormalParameterList formalParameterList;

  FactoryConstructorSignature(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'factory');
    lexer.advance();

    identifier1 = Identifier(lexer);

    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.PERIOD);
      lexer.advance();

      identifier2 = Identifier(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    formalParameterList = FormalParameterList(lexer);
  }

  String toString() {
    return 'factory $identifier1' +
        (identifier2 != null ? '.$identifier2' : '') +
        '$formalParameterList';
  }
}

/// <functionSignature> ::=
/// <metadata> <type>? <identifier> <formalParameterPart>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 17
class FunctionSignature extends ParserType {
  MetaData metaData;
  Type type;
  Identifier identifier;
  FormalParameterPart formalParameterPart;

  FunctionSignature(Lexer lexer) {
    metaData = MetaData(lexer);

    int snapShot = lexer.currentIndex;
    try {
      type = Type(lexer);

      // This is for the example:
      // > const x = 0xfffff
      // Where there is no 'var' because it is 'const' or 'final'
      checkNotType(lexer, TType.EQUALS);

      // This is for the examples:
      // > test() {}
      // > test<someType>() {}
      // Where there is no type because it is dynamic
      checkNotType(lexer, TType.LEFT_ANGLE);
      checkNotType(lexer, TType.LEFT_PAREN);
    } on WrongTokenType {
      lexer.setPos(snapShot);
      type = null;
    }

    identifier = Identifier(lexer);
    formalParameterPart = FormalParameterPart(lexer);
  }

  String toString() {
    return '$metaData' +
        (type != null ? '$type ' : '') +
        '$identifier$formalParameterPart';
  }
}

/// <getterSignature> ::= <type>? get <identifier>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 34
class GetterSignature extends ParserType {
  Type type;
  Identifier identifier;

  GetterSignature(Lexer lexer) {
    int snapShot = lexer.currentIndex;
    try {
      type = Type(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'get');
    lexer.advance();

    identifier = Identifier(lexer);
  }

  String toString() {
    return (type != null ? '$type ' : '') + 'get $identifier';
  }
}

/// <setterSignature> ::= <type>? set <identifier> <formalParameterList>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 35
class SetterSignature extends ParserType {
  Type type;
  Identifier identifier;
  FormalParameterList formalParameterList;

  SetterSignature(Lexer lexer) {
    int snapShot = lexer.currentIndex;
    try {
      type = Type(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'set');
    lexer.advance();

    identifier = Identifier(lexer);
    formalParameterList = FormalParameterList(lexer);
  }

  String toString() {
    return (type != null ? '$type ' : '') +
        'set $identifier$formalParameterList';
  }
}

/// <operatorSignature> ::=
/// <type>? operator <operator> <formalParameterList>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 29
class OperatorSignature extends ParserType {
  Type type;
  Operator operator_;
  FormalParameterList formalParameterList;

  OperatorSignature(Lexer lexer) {
    int snapShot = lexer.currentIndex;
    try {
      type = Type(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    operator_ = Operator(lexer);
    formalParameterList = FormalParameterList(lexer);
  }

  String toString() {
    return (type != null ? '$type ' : '') +
        'operator $operator_$formalParameterList';
  }
}

/// <constantConstructorSignature> ::= const <qualified> <formalParameterList>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 45
class ConstantConstructorSignature extends ParserType {
  Qualified qualified;
  FormalParameterList formalParameterList;

  ConstantConstructorSignature(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'const');
    lexer.advance();

    qualified = Qualified(lexer);
    formalParameterList = FormalParameterList(lexer);
  }

  String toString() {
    return 'const $qualified$formalParameterList';
  }
}

/// <redirection> ::= ':' this ('.' <identifier>)? <arguments>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 39
class Redirection extends ParserType {
  Identifier identifier;
  Arguments arguments;

  Redirection(Lexer lexer) {
    checkType(lexer, TType.COLON);
    lexer.advance();

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'this');
    lexer.advance();

    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.PERIOD);
      lexer.advance();

      identifier = Identifier(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    arguments = Arguments(lexer);
  }

  String toString() {
    return ': this' +
        (identifier != null ? '.$identifier' : '') +
        ' $arguments';
  }
}
