import '../ParserType.dart';
import '../ParserExceptions.dart';
import '../lexer/Lexer.dart';
import '../sub_types/Identifiers.dart';
import '../lists/ParameterTypeList.dart';
import '../lists/TypeParameters.dart';
import '../lists/TypeList.dart';
import '../core/Type.dart';

/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 14

/// <finalConstVarOrType> ::= final <type>?
/// | const <type>?
/// | <varOrType>
class FinalConstVarOrType extends ParserType {
  int caseNum;
  Type type;
  VarOrType varOrType;

  FinalConstVarOrType(Lexer lexer) {
    // print ('Starting FinalConstVarOrType');
    int attempt, maxAttempts = 3;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
          case 2:
            _case2(lexer);
            break;
        }

        // print ('Completed FinalConstVarOrType $attempt');
        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType(
          'Tried all options in FinalConstVarOrType: ALL FAILED');
    }
  }

  /// final <type>?
  void _case0(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'final');
    lexer.advance();

    int snapShot = lexer.currentIndex;
    try {
      type = Type(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    caseNum = 0;
  }

  /// const <type>?
  void _case1(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'const');
    lexer.advance();

    int snapShot = lexer.currentIndex;
    try {
      type = Type(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    caseNum = 1;
  }

  /// <varOrType>
  void _case2(Lexer lexer) {
    varOrType = VarOrType(lexer);
    caseNum = 2;
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return 'final' + (type != null ? ' $type' : '');
        break;
      case 1:
        return 'const' + (type != null ? ' $type' : '');
        break;
      case 2:
        return '$varOrType';
        break;
    }
    return '';
  }
}

/// <varOrType> ::= var | <type>
class VarOrType extends ParserType {
  int caseNum;
  Type type;

  VarOrType(Lexer lexer) {
    // print ('Starting VarOrType');
    int attempt, maxAttempts = 2;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
        }

        // print ('Completed VarOrType $attempt');
        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType('Tried all options in VarOrType: ALL FAILED');
    }
  }

  /// var
  void _case0(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'var');
    lexer.advance();

    caseNum = 0;
  }

  /// <type>
  void _case1(Lexer lexer) {
    type = Type(lexer);
    caseNum = 1;
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return 'var';
        break;
      case 1:
        return '$type';
    }
    return '';
  }
}

// --------------------------------------------------------------------------

/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 176

/// <typeNotFunction> ::= <typeNotVoidNotFunction> | void
class TypeNotFunction extends ParserType {
  int caseNum;
  TypeNotVoidNotFunction typeNotVoidNotFunction;

  TypeNotFunction(Lexer lexer) {
    // print ('Starting TypeNotFunction');
    int attempt, maxAttempts = 2;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
        }

        // print ('Completed TypeNotFunction $attempt');
        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType('Tried all options in TypeNotFunction: ALL FAILED');
    }
  }

  /// void
  void _case0(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'void');
    lexer.advance();

    caseNum = 0;
  }

  /// <typeNotVoidNotFunction>
  void _case1(Lexer lexer) {
    typeNotVoidNotFunction = TypeNotVoidNotFunction(lexer);
    caseNum = 1;
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return 'void';
        break;
      case 1:
        return '$typeNotVoidNotFunction';
    }
    return '';
  }
}

/// <typeNotVoidNotFunction> ::= <typeName> <typeArguments>? | Function
class TypeNotVoidNotFunction extends ParserType {
  int caseNum;
  TypeName typeName;
  TypeArguments typeArguments;

  TypeNotVoidNotFunction(Lexer lexer) {
    // print ('Starting TypeNotVoidNotFunction');
    int attempt, maxAttempts = 2;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
        }

        // print ('Completed TypeNotVoidNotFunction $attempt');
        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType(
          'Tried all options in TypeNotVoidNotFunction: ALL FAILED');
    }
  }

  /// Function
  void _case0(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'Function');
    lexer.advance();

    caseNum = 0;
  }

  /// <typeName> <typeArguments>?
  void _case1(Lexer lexer) {
    typeName = TypeName(lexer);

    int snapShot = lexer.currentIndex;
    try {
      typeArguments = TypeArguments(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    caseNum = 1;
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return 'Function';
        break;
      case 1:
        return '$typeName' + (typeArguments != null ? '$typeArguments' : '');
        break;
    }
    return '';
  }
}

/// <typeName> ::= <typeIdentifier> ('.' <typeIdentifier>)?
class TypeName extends ParserType {
  TypeIdentifier typeIdentifier1;
  TypeIdentifier typeIdentifier2;

  TypeName(Lexer lexer) {
    typeIdentifier1 = TypeIdentifier(lexer);

    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.PERIOD);
      lexer.advance();
      typeIdentifier2 = TypeIdentifier(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }
  }

  String toString() {
    return '$typeIdentifier1' +
        (typeIdentifier2 != null ? '.$typeIdentifier2' : '');
  }
}

/// <typeArguments> ::= '<' <typeList> '>'
class TypeArguments extends ParserType {
  TypeList typeList;

  TypeArguments(Lexer lexer) {
    // print ('Starting TypeArguments');
    // print (lexer.currentToken);
    // print (lexer.nextToken);

    checkType(lexer, TType.LEFT_ANGLE);
    lexer.advance();

    typeList = TypeList(lexer);

    checkType(lexer, TType.RIGHT_ANGLE);
    lexer.advance();

    // print ('Completed TypeArguments');
  }

  String toString() {
    return '<$typeList>';
  }
}

/// <typeNotVoid> ::= <functionType> | <typeNotVoidNotFunction>
class TypeNotVoid extends ParserType {
  FunctionType functionType;
  TypeNotVoidNotFunction typeNotVoidNotFunction;

  TypeNotVoid(Lexer lexer) {
    int snapShot = lexer.currentIndex;
    try {
      functionType = FunctionType(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);

      typeNotVoidNotFunction = TypeNotVoidNotFunction(lexer);
    }
  }

  String toString() {
    return (functionType != null ? '$functionType' : '$typeNotVoidNotFunction');
  }
}

/// <functionType> ::= <functionTypeTails> | <typeNotFunction> <functionTypeTails>
class FunctionType extends ParserType {
  FunctionTypeTails functionTypeTails;
  TypeNotFunction typeNotFunction;

  FunctionType(Lexer lexer) {
    int snapShot = lexer.currentIndex;
    try {
      functionTypeTails = FunctionTypeTails(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);

      typeNotFunction = TypeNotFunction(lexer);
      functionTypeTails = FunctionTypeTails(lexer);
    }
  }
}

/// <functionTypeTails> ::= <functionTypeTail> <functionTypeTails> | <functionTypeTail>
class FunctionTypeTails extends ParserType {
  FunctionTypeTail functionTypeTail;
  FunctionTypeTails functionTypeTails;

  FunctionTypeTails(Lexer lexer) {
    int snapShot = lexer.currentIndex;
    try {
      functionTypeTail = FunctionTypeTail(lexer);
      functionTypeTails = FunctionTypeTails(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);

      functionTypeTail = FunctionTypeTail(lexer);
    }
  }
}

/// <functionTypeTail> ::= Function <typeParameters>? <parameterTypeList>
class FunctionTypeTail extends ParserType {
  TypeParameters typeParameters;
  ParameterTypeList parameterTypeList;

  FunctionTypeTail(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'Function');
    lexer.advance();

    int snapShot = lexer.currentIndex;
    try {
      typeParameters = TypeParameters(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    parameterTypeList = ParameterTypeList(lexer);
  }
}
