import '../ParserType.dart';
import '../ParserExceptions.dart';
import '../lexer/Lexer.dart';
import '../core/Identifier.dart';
import '../main/MetaData.dart';
import 'Combinator.dart';
import 'Uries.dart';

/// <libraryImport> ::= <metadata> <importSpecification>
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 167
class LibraryImport extends ParserType {
  MetaData metaData;
  ImportSpecification importSpecification;

  LibraryImport(Lexer lexer) {
    metaData = MetaData(lexer);
    importSpecification = ImportSpecification(lexer);
  }

  String toString() {
    return '$metaData$importSpecification';
  }
}

/// <importSpecification> ::= import <configurableUri> (as <identifier>)? <combinator>* ';'
/// | import <uri> deferred as <identifier> <combinator>* ';'
class ImportSpecification extends ParserType {
  int caseNum;
  ConfigurableUri configurableUri;
  Identifier identifier;
  List<Combinator> combinators;
  Uri uri;

  ImportSpecification(Lexer lexer) {
    combinators = [];

    int attempt, maxAttempts = 2;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType(
          'Tried all options in ImportSpecification: ALL FAILED');
    }
  }

  /// import <configurableUri> (as <identifier>)? <combinator>* ';'
  void _case0(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'import');
    lexer.advance();

    configurableUri = ConfigurableUri(lexer);

    int snapShot = lexer.currentIndex;
    try {
      checkType(lexer, TType.KEYWORD);
      checkVal(lexer, 'as');
      lexer.advance();

      identifier = Identifier(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    try {
      while (true) {
        snapShot = lexer.currentIndex;
        combinators.add(Combinator(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    checkType(lexer, TType.SEMICOLON);
    lexer.advance();

    caseNum = 0;
  }

  /// import <uri> deferred as <identifier> <combinator>* ';'
  void _case1(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'import');
    lexer.advance();

    uri = Uri(lexer);

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'deferred');
    lexer.advance();

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'as');
    lexer.advance();

    identifier = Identifier(lexer);

    int snapShot;
    try {
      while (true) {
        snapShot = lexer.currentIndex;
        combinators.add(Combinator(lexer));
      }
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    checkType(lexer, TType.SEMICOLON);
    lexer.advance();

    caseNum = 1;
  }

  String toString() {
    switch (caseNum) {
      case 0:
        String rtn = 'import $configurableUri ' +
            (identifier != null ? 'as $identifier ' : '');
        combinators.forEach((c) => rtn += '$c ');
        return rtn.trimRight() + ';';
        break;
      case 1:
        String rtn = 'import $uri deferred as $identifier';
        combinators.forEach((c) => rtn += ' $c');
        return rtn + ';';
        break;
    }
    return '';
  }
}
