import '../ParserType.dart';
import '../ParserExceptions.dart';
import '../lexer/Lexer.dart';
import '../core/Identifier.dart';
import '../core/Type.dart';
import '../main/VariableDeclaration.dart';
import '../main/ClassDefinition.dart';
import '../main/EnumType.dart';
import '../lists/StaticFinalDeclarationList.dart';
import '../lists/FormalParameterList.dart';
import '../sub_types/FunctionBody.dart';
import '../sub_types/Signatures.dart';
import '../sub_types/TypeAlias.dart';

/// <topLevelDefinition> ::= <classDefinition>
/// | <enumType>
/// | <typeAlias>
/// | external? <functionSignature> ';'
/// | external? <getterSignature> ';'
/// | external? <setterSignature> ';'
/// | <functionSignature> <functionBody>
/// | <type>? get <identifier> <functionBody>
/// | <type>? set <identifier> <formalParameterList> <functionBody>
/// | (final | const) <type> <staticFinalDeclarationList> ';'
/// | <variableDeclaration> ';'
///
/// See https://dart.dev/guides/language/specifications/DartLangSpec-v2.2.pdf
/// Page 165-166
class TopLevelDefinition extends ParserType {
  int caseNum;
  StaticFinalDeclarationList staticFinalDeclarationList;
  FormalParameterList formalParameterList;
  VariableDeclaration variableDeclaration;
  FunctionSignature functionSignature;
  GetterSignature getterSignature;
  SetterSignature setterSignature;
  ClassDefinition classDefinition;
  FunctionBody functionBody;
  Identifier identifier;
  TypeAlias typeAlias;
  EnumType enumType;
  Type type;

  bool isExternal;
  bool isConst;

  TopLevelDefinition(Lexer lexer) {
    isExternal = false;
    isConst = false;

    int attempt, maxAttempts = 11;
    int snapShot = lexer.currentIndex;
    for (attempt = 0; attempt < maxAttempts; ++attempt) {
      try {
        switch (attempt) {
          case 0:
            _case0(lexer);
            break;
          case 1:
            _case1(lexer);
            break;
          case 2:
            _case2(lexer);
            break;
          case 3:
            _case3(lexer);
            break;
          case 4:
            _case4(lexer);
            break;
          case 5:
            _case5(lexer);
            break;
          case 6:
            _case6(lexer);
            break;
          case 7:
            _case7(lexer);
            break;
          case 8:
            _case8(lexer);
            break;
          case 9:
            _case9(lexer);
            break;
          case 10:
            _case10(lexer);
            break;
        }

        break;
      } on WrongTokenType {
        lexer.setPos(snapShot);
      }
    }

    if (attempt == maxAttempts) {
      throw WrongTokenType(
          'Tried all options in TopLevelDefinition: ALL FAILED');
    }
  }

  /// <classDefinition>
  void _case0(Lexer lexer) {
    classDefinition = ClassDefinition(lexer);
    caseNum = 0;
  }

  /// <enumType>
  void _case1(Lexer lexer) {
    enumType = EnumType(lexer);
    caseNum = 1;
  }

  /// <typeAlias>
  void _case2(Lexer lexer) {
    typeAlias = TypeAlias(lexer);
    caseNum = 2;
  }

  /// external? <functionSignature> ';'
  void _case3(Lexer lexer) {
    _checkExternal(lexer);
    functionSignature = FunctionSignature(lexer);

    checkType(lexer, TType.SEMICOLON);
    lexer.advance();

    caseNum = 3;
  }

  /// external? <getterSignature> ';'
  void _case4(Lexer lexer) {
    _checkExternal(lexer);
    getterSignature = GetterSignature(lexer);

    checkType(lexer, TType.SEMICOLON);
    lexer.advance();

    caseNum = 4;
  }

  /// external? <setterSignature> ';'
  void _case5(Lexer lexer) {
    _checkExternal(lexer);
    setterSignature = SetterSignature(lexer);

    checkType(lexer, TType.SEMICOLON);
    lexer.advance();

    caseNum = 5;
  }

  /// <functionSignature> <functionBody>
  void _case6(Lexer lexer) {
    functionSignature = FunctionSignature(lexer);
    functionBody = FunctionBody(lexer);
    caseNum = 6;
  }

  /// <type>? get <identifier> <functionBody>
  void _case7(Lexer lexer) {
    int snapShot = lexer.currentIndex;
    try {
      type = Type(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'get');
    lexer.advance();

    identifier = Identifier(lexer);
    functionBody = FunctionBody(lexer);

    caseNum = 7;
  }

  /// <type>? set <identifier> <formalParameterList> <functionBody>
  void _case8(Lexer lexer) {
    int snapShot = lexer.currentIndex;
    try {
      type = Type(lexer);
    } on WrongTokenType {
      lexer.setPos(snapShot);
    }

    checkType(lexer, TType.KEYWORD);
    checkVal(lexer, 'set');
    lexer.advance();

    identifier = Identifier(lexer);
    formalParameterList = FormalParameterList(lexer);
    functionBody = FunctionBody(lexer);

    caseNum = 8;
  }

  /// (final | const) <type> <staticFinalDeclarationList> ';'
  void _case9(Lexer lexer) {
    checkType(lexer, TType.KEYWORD);
    if (isVal(lexer, 'final')) {
      isConst = false;
    } else if (isVal(lexer, 'const')) {
      isConst = true;
    } else {
      throw WrongTokenType('Must be either final or const');
    }
    lexer.advance();

    type = Type(lexer);
    staticFinalDeclarationList = StaticFinalDeclarationList(lexer);

    checkType(lexer, TType.SEMICOLON);
    lexer.advance();

    caseNum = 9;
  }

  /// <variableDeclaration> ';'
  void _case10(Lexer lexer) {
    variableDeclaration = VariableDeclaration(lexer);

    checkType(lexer, TType.SEMICOLON);
    lexer.advance();

    caseNum = 10;
  }

  void _checkExternal(Lexer lexer) {
    if (isType(lexer, TType.KEYWORD) && isVal(lexer, 'external')) {
      isExternal = true;
      lexer.advance();
    }
  }

  String toString() {
    switch (caseNum) {
      case 0:
        return '$classDefinition';
        break;
      case 1:
        return '$enumType';
        break;
      case 2:
        return '$typeAlias';
        break;
      case 3:
        return (isExternal ? 'external ' : '') + '$functionSignature;';
        break;
      case 4:
        return (isExternal ? 'external ' : '') + '$getterSignature;';
        break;
      case 5:
        return (isExternal ? 'external ' : '') + '$setterSignature;';
        break;
      case 6:
        return '$functionSignature $functionBody';
        break;
      case 7:
        return (type != null ? '$type ' : '') + 'get $identifier $functionBody';
        break;
      case 8:
        return (type != null ? '$type ' : '') +
            'set $identifier $formalParameterList $functionBody';
        break;
      case 9:
        return (isConst ? 'const' : 'final') +
            ' $type $staticFinalDeclarationList;';
        break;
      case 10:
        return '$variableDeclaration;';
        break;
    }
    return '';
  }
}
